import re
from inspect import _ParameterKind
from pydantic import BaseModel
from dataclasses import dataclass
from typing import List, Tuple, Optional, Any, Dict, TYPE_CHECKING

from bridgic.core.automa.worker import Worker
from bridgic.core.types._error import WorkerArgsInjectionError
from bridgic.core.utils._args_map import safely_map_args

if TYPE_CHECKING:
    from bridgic.core.automa._graph_automa import GraphAutoma


class InjectorNone: 
    """
    Marker object for Injector.inject() when the default value is None.
    """
    ...

class ArgsDescriptor:
    """
    A descriptor for arguments that can be injected.
    """
    ...

@dataclass
class From(ArgsDescriptor):
    """
    Implementing arguments injection for worker parameters with default value.

    When a worker needs the output of another worker but does not directly depend on 
    it in execution, you can use `From` to declare an arguments injection in 
    its parameters.

    Attributes
    ----------
    key : str
        The key of the worker to inject arguments from.
    default : Optional[Any]
        The default value of the arguments.

    Examples
    --------
    ```python
    class MyAutoma(GraphAutoma):
        @worker(is_start=True)
        def worker_0(self, user_input: int) -> int:
            return user_input + 1
        
        @worker(dependencies=["worker_0"])
        def worker_1(self, worker_0_output: int) -> int:
            return worker_0_output + 1
        
        @worker(dependencies=["worker_1"], is_output=True)
        def worker_2(self, worker_1_output: int, worker_0_output: int = From("worker_0", 1)) -> int:
            # needs the output of worker_0 but does not directly depend on it in execution
            print(f'worker_0_output: {worker_0_output}')
            return worker_1_output + 1
    ```

    Returns
    -------
    Any
        The output of the worker specified by the key.

    Raises
    ------
    WorkerArgsInjectionError
        If the worker specified by the key does not exist and no default value is set.

    Note:
    ------
    1. Can set a default value for a `From` declaration, which will be returned when the specified worker does not exist.
    2. Will raise `WorkerArgsInjectionError` if the worker specified by the key does not exist and no default value is set.
    """
    key: str
    default: Optional[Any] = InjectorNone()

def resolve_from(dep: From, worker_output: Dict[str, Any]) -> Any:
    inject_res = worker_output.get(dep.key, dep.default)
    if isinstance(inject_res, InjectorNone):
        raise WorkerArgsInjectionError(
            f"the worker: `{dep.key}` is not found in the automa or `{dep.key}` is already removed. "
            "You may need to set the default value of the parameter to a `From` instance with the key of the worker."
        )
    return inject_res

@dataclass
class System(ArgsDescriptor):
    """
    Implementing system-level arguments injection for worker parameters.
        
    System provides access to automa-level resources and context through arguments 
    injection. It supports pattern matching for different types of system resources.

    Attributes
    ----------
    key : str
        The system resource key to inject. Supported keys:
        - "runtime_context": Runtime context for data persistence across worker executions.
        - "automa": Current automa instance.
        - "automa:worker_key": Sub-automa instance in current automa.

    Examples
    --------
    ```python
    def worker_1(x: int, current_automa = System("automa")) -> int:
        # Access current automa instance
        current_automa.add_worker(
            key="sub_automa",
            worker=SubAutoma(),
            dependencies=["worker_1"]
        )
        return x + 1

    class SubAutoma(GraphAutoma):
        @worker(is_start=True)
        def worker_0(self, user_input: int) -> int:
            return user_input + 1

    class MyAutoma(GraphAutoma):
        @worker(is_start=True)
        def worker_0(self, user_input: int, rtx = System("runtime_context")) -> int:
            # Access runtime context for data persistence
            local_space = self.get_local_space(rtx)
            count = local_space.get("count", 0)
            local_space["count"] = count + 1

            self.add_func_as_worker(
                key="worker_1",
                func=worker_1,
                dependencies=["worker_0"]
            )

            return user_input + count
            
        @worker(dependencies=["worker_1"])
        def worker_2(self, worker_1_output: int, sub_automa = System("automa:sub_automa")) -> int:
            # Access sub-automa from worker_1
            sub_automa.add_worker(
                key="worker_3",
                worker=SubAutoma(),
                dependencies=["worker_2"],
                is_output=True,
            )
            return worker_1_output + 1
    ```

    Returns
    -------
    Any
        The system resource specified by the key:
        - RuntimeContext: For "runtime_context"
        - AutomaInstance: For current automa instance or a sub-automa instance from the current automa.

    Raises
    ------
    WorkerArgsInjectionError
        - If the key pattern is not supported.
        - If the specified resource does not exist.
        - If the specified resource is not an Automa.

    Note
    ----
    1. "runtime_context" provides a `RuntimeContext` instance for data persistence
    2. "automa" provides access to the current automa instance
    3. "automa:worker_key" provides access to a sub-automa from the specified worker key
    """
    key: str
    
    def __post_init__(self):
        allowed_patterns = [
            r"^runtime_context$",
            r"^automa:.*$",
            r"^automa$",
        ]
        
        if not any(re.match(pattern, self.key) for pattern in allowed_patterns):
            raise WorkerArgsInjectionError(
                f"Key '{self.key}' is not supported. Supported keys: \n"
                f"- `runtime_context`: a context for data persistence of the current worker.\n"
                f"- `automa:<worker_key>`: a sub-automa in current automa.\n"
                f"- `automa`: the current automa instance.\n"
            )

JSON_SCHEMA_IGNORE_ARG_TYPES = (System, From)

def resolve_system(dep: System, current_worker_key: str, worker_dict: Dict[str, Worker], current_automa: "GraphAutoma") -> Any:
    if dep.key == "runtime_context":
        return RuntimeContext(worker_key=current_worker_key)
    elif dep.key.startswith("automa:"):
        worker_key = dep.key[7:]

        inject_res = worker_dict.get(worker_key, InjectorNone())
        if isinstance(inject_res, InjectorNone):
            raise WorkerArgsInjectionError(
                f"the sub-atoma: `{dep.key}` is not found in current automa. "
            )

        if not inject_res.is_automa():
            raise WorkerArgsInjectionError(
                f"the `{dep.key}` instance is not an Automa. "
            )
        
        return inject_res.get_decorated_worker()
    elif dep.key == "automa":
        return current_automa

class RuntimeContext(BaseModel):
    worker_key: str

class WorkerInjector:
    """
    Worker Dependency injection container for resolving dependency data injection of workers.

    This class manages workers for dependency injection, allowing you to inject 
    dependencies into function parameters based on their default values. 
    It is used with the `From` and `System` classes to implement dependency injection of workers.
    """
    
    def dump_to_dict(self) -> Dict[str, Any]:
        """
        Serialize WorkerInjector to dictionary.
        Since WorkerInjector is stateless, we just return an empty dict.
        """
        return {"type": "WorkerInjector"}
    
    @classmethod
    def load_from_dict(cls, data: Dict[str, Any]) -> "WorkerInjector":
        """
        Deserialize WorkerInjector from dictionary.
        Since WorkerInjector is stateless, we just return a new instance.
        """
        return cls()

    def inject(
        self, 
        current_worker_key: str,
        current_worker_sig: Dict[_ParameterKind, List],
        current_automa: "GraphAutoma"
    ) -> Any:
        """
        Inject dependencies into parameters whose default value is a `From` or `System`.

        Parameters
        ----------
        current_worker_key : str
            The key of the current worker being processed.
        current_worker_sig : Dict[_ParameterKind, List]
            Dictionary mapping parameters to their signature information of the current worker.
        current_automa : GraphAutoma
            The current automa instance.
            
        Returns
        -------
        Tuple[Tuple[Any, ...], Dict[str, Any]]
            A tuple containing the keyword arguments for the current worker to be injected.
        """
        worker_dict = current_automa._workers
        worker_output = current_automa._worker_output

        param_list = [
            param
            for _, param_list in current_worker_sig.items()
            for param in param_list
        ]

        from_inject_kwargs = {}
        system_inject_kwargs = {}
        for name, default_value in param_list:
            if isinstance(default_value, From):
                value = resolve_from(default_value, worker_output)
                from_inject_kwargs[name] = value
            elif isinstance(default_value, System):
                value = resolve_system(default_value, current_worker_key, worker_dict, current_automa)
                system_inject_kwargs[name] = value

        current_kwargs = {
            **from_inject_kwargs,
            **system_inject_kwargs
        }
        return (), current_kwargs




"""
The Agentic Types module defines foundational data structures for agentic systems.

This module defines several important type definitions, such as `ToolSpec` and 
`ChatMessage`, which are designed to be "model-neutral" as much as possible, 
allowing developers to build agentic systems using different models.
"""

from bridgic.core.agentic.types._chat_message import *

__all__ = [
    "Function",
    "FunctionToolCall",
    "FunctionToolSpec",
    "ToolSpec",
    "ToolMessage",
    "ChatMessage",
    "SystemMessage",
    "UserTextMessage",
    "AssistantTextMessage",
]


"""
The Model module provides core abstraction entities for LLMs (Large Language Models).

This module defines core abstraction entities for interacting with models, providing 
foundational type abstractions for different model implementations.
"""

from bridgic.core.model._base_llm import BaseLlm

__all__ = [
    "BaseLlm",
]


from pydantic import BaseModel
from typing import Optional, Any, Generator, AsyncGenerator

from bridgic.core.model.types._message import Message, MessageChunk

class Response(BaseModel):
    """
    LLM response container for model outputs.

    Represents the complete response from a language model, containing both
    the message content and the raw response data from the underlying model 
    provider.

    Attributes
    ----------
    message : Optional[Message]
        The structured message containing the model's response content.
    raw : Optional[Any]
        Raw response data from the LLM provider for debugging or custom processing.
    """
    message: Optional[Message] = None
    raw: Optional[Any] = None

StreamResponse = Generator[MessageChunk, None, None]
AsyncStreamResponse = AsyncGenerator[MessageChunk, None]



from typing import List, Any, Protocol, Literal, Union, Dict, Type, ClassVar
from pydantic import BaseModel, Field

from bridgic.core.model.types import Message

class PydanticModel(BaseModel):
    """
    A constraint defined as a Pydantic model for structured LLM output.
    """
    constraint_type: Literal["pydantic_model"] = "pydantic_model"
    """The type of the constraint, in this case `pydantic_model`."""
    model: Type[BaseModel] = Field(..., description="Model type of the PydanticModel constraint.")
    """The Pydantic model type of the constraint."""

class JsonSchema(BaseModel):
    """
    A constraint defined as a JSON schema for structured LLM output.
    """
    constraint_type: Literal["json_schema"] = "json_schema"
    """The type of the constraint, in this case `json_schema`."""
    schema_dict: Dict[str, Any] = Field(..., description="Schema of the JsonSchema constraint.")
    """The JSON schema of the constraint."""

class Regex(BaseModel):
    """
    A constraint defined as a regular expression for structured LLM output.
    """
    constraint_type: Literal["regex"] = "regex"
    """The type of the constraint, in this case `regex`."""
    pattern: str = Field(..., description="Pattern of the Regex constraint.")
    """The regular expression of the constraint."""

class RegexPattern:
    """Constants that define some common regular expressions for structured LLM output."""
    INTEGER: ClassVar[Regex] = Regex(pattern=r"-?\d+")
    """A regular expression for integers."""
    FLOAT = Regex(pattern=r"-?(?:\d+\.\d+|\d+\.|\.\d+|\d+)([eE][-+]?\d+)?")
    """A regular expression for floats."""
    DATE: ClassVar[Regex] = Regex(pattern=r"\d{4}-\d{2}-\d{2}")
    """A regular expression for dates."""
    TIME: ClassVar[Regex] = Regex(pattern=r"(?:[01]\d|2[0-3]):[0-5]\d:[0-5]\d(?:\.\d+)?")
    """A regular expression for times."""
    DATE_TIME_ISO_8601: ClassVar[Regex] = Regex(pattern=rf"{DATE.pattern}T{TIME.pattern}(?:Z|[+-](?:[01]\d|2[0-3]):[0-5]\d)?")
    """A regular expression for date-time in ISO 8601 format."""
    IP_V4_ADDRESS: ClassVar[Regex] = Regex(pattern=r"(?:(?:25[0-5]|2[0-4]\d|1\d{2}|[1-9]?\d)\.){3}(?:25[0-5]|2[0-4]\d|1\d{2}|[1-9]?\d)")
    """A regular expression for IPv4 addresses."""
    IP_V6_ADDRESS: ClassVar[Regex] = Regex(pattern=r"([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}")
    """A regular expression for IPv6 addresses."""
    EMAIL: ClassVar[Regex] = Regex(pattern=r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+")
    """A regular expression for email addresses."""

class Choice(BaseModel):
    """
    A constraint defined as a predefined set of choices for structured LLM output.
    """
    constraint_type: Literal["choice"] = "choice"
    """The type of the constraint, in this case `choice`."""
    choices: List[str] = Field(..., description="Choices of the choice constraint.")
    """The choices of the constraint."""

class EbnfGrammar(BaseModel):
    """
    A constraint defined as an EBNF grammar for structured LLM output.
    """
    constraint_type: Literal["ebnf_grammar"] = "ebnf_grammar"
    """The type of the constraint, in this case `ebnf_grammar`."""
    syntax: str = Field(..., description="Syntax of the EBNF grammar constraint.")
    """The syntax of the EBNF grammar constraint."""

class LarkGrammar(BaseModel):
    """
    A constraint defined as a Lark grammar for structured LLM output.
    """
    constraint_type: Literal["lark_grammar"] = "lark_grammar"
    """The type of the constraint, in this case `lark_grammar`."""
    syntax: str = Field(..., description="Syntax of the Lark grammar constraint.")
    """The syntax of the Lark grammar constraint."""

Constraint = Union[PydanticModel, JsonSchema, EbnfGrammar, LarkGrammar, Regex, Choice]
"""The constraint type for structured LLM output."""

class StructuredOutput(Protocol):
    """
    Protocol for LLM providers that support structured output generation.

    StructuredOutput defines the interface for language models that can generate 
    responses in specific formats according to given constraints. This protocol 
    enables controlled output generation for various data structures and formats.

    Methods
    -------
    structured_output
        Synchronous method for generating structured output based on constraints.
    astructured_output
        Asynchronous method for generating structured output based on constraints.

    Notes
    ----
    1. Both synchronous and asynchronous methods must be implemented
    2. Supported constraint types depend on the specific LLM provider implementation
    3. Output format is determined by the constraint type provided
    4. Common constraint types include PydanticModel, JsonSchema, Regex, Choice, etc.
    """

    def structured_output(
        self,
        messages: List[Message],
        constraint: Constraint,
        **kwargs,
    ) -> Any:
        """
        Generate structured output based on conversation context and constraints.

        Parameters
        ----------
        messages : List[Message]
            The conversation history and current context.
        constraint : Constraint
            The output format constraint. Supported types:

            - PydanticModel: Output as Pydantic model instance
            - JsonSchema: Output as JSON matching the schema
            - Regex: Output matching the regex pattern
            - Choice: Output from predefined choices
            - EbnfGrammar: Output following EBNF grammar rules
            - LarkGrammar: Output following Lark grammar rules
        **kwargs
            Additional keyword arguments for output generation configuration.

        Returns
        -------
        Any
            The structured output matching the specified constraint format.
        """
        ...

    async def astructured_output(
        self,
        messages: List[Message],
        constraint: Constraint,
        **kwargs,
    ) -> Any:
        """
        Asynchronously generate structured output based on conversation context and constraints.

        Parameters
        ----------
        messages : List[Message]
            The conversation history and current context.
        constraint : Constraint
            The output format constraint. Supported types:

            - PydanticModel: Output as Pydantic model instance
            - JsonSchema: Output as JSON matching the schema
            - Regex: Output matching the regex pattern
            - Choice: Output from predefined choices
            - EbnfGrammar: Output following EBNF grammar rules
            - LarkGrammar: Output following Lark grammar rules
        **kwargs
            Additional keyword arguments for output generation configuration.

        Returns
        -------
        Any
            The structured output matching the specified constraint format.
        """
        ...


from typing import Any, Dict, List, Tuple, Optional
from typing_extensions import override

from bridgic.core.automa.worker import Worker
from bridgic.core.model.types import Message, Tool, ToolCall
from bridgic.core.model.protocols import ToolSelection
from bridgic.core.agentic.types._chat_message import ChatMessage
from bridgic.core.prompt.utils._prompt_utils import transform_chat_message_to_llm_message

class ToolSelectionWorker(Worker):
    """
    A worker that calls an LLM to select tools and/or generate a response.
    """

    # Note: the ToolSelection LLM instance need support serialization and deserialization.
    _tool_selection_llm: ToolSelection
    """The LLM to be used for tool selection."""

    def __init__(self, tool_selection_llm: ToolSelection):
        """
        Parameters
        ----------
        tool_selection_llm: ToolSelect
            The LLM to be used for tool selection.
        """
        super().__init__()
        self._tool_selection_llm = tool_selection_llm

    async def arun(
        self,
        messages: List[ChatMessage],
        tools: List[Tool],
    ) -> Tuple[List[ToolCall], Optional[str]]:
        """
        Run the worker.

        Parameters
        ----------
        messages: List[ChatMessage]
            The messages to send to the LLM.
        tools: List[Tool]
            The tool list for the LLM to select from.

        Returns
        -------
        Tuple[List[ToolCall], Optional[str]]
            * The first element is a list of `ToolCall` that the LLM selected.
            * The second element is the text response from the LLM.
        """
        # Validate and transform the input messages and tools to the format expected by the LLM.
        llm_messages: List[Message] = []
        for message in messages:
            llm_messages.append(transform_chat_message_to_llm_message(message))
        # print(f"\n******* ToolSelectionWorker.arun *******\n")
        # print(f"messages: {llm_messages}")
        # print(f"tools: {tools}")
        tool_calls, llm_response = await self._tool_selection_llm.aselect_tool(
            messages=llm_messages, 
            tools=tools, 
        )
        return tool_calls, llm_response

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()
        state_dict["tool_selection_llm"] = self._tool_selection_llm
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)
        self._tool_selection_llm = state_dict["tool_selection_llm"]


"""
The Agentic Workers module provides specialized implementation of Worker for agentic systems.

This module provides specialized Worker implementations for specific functions to support 
building Agentic systems with complex capabilities.
"""

from ._tool_selection import ToolSelectionWorker

__all__ = [
    "ToolSelectionWorker"
]


import inspect
import importlib
import enum

from typing import Callable, List, Dict, Any, Tuple
from typing_extensions import get_overloads, overload
from bridgic.core.utils._collection import deep_hash

_marked_overloads: Dict[str, Dict[str, Any]] = {}

def mark_overload(key: str, value: Any) -> Callable:
    """
    A decorator to mark an overload function with a key and value. It is useful 
    when you need to mark a function as overloaded and add more information to it.
    """
    def wrapper(func: Callable):
        func_key = f"{func.__module__}.{func.__qualname__}"
        params_key = hash_kw_default_params(func)
        func_params_key = f"{func_key}::{params_key}"
        if func_params_key not in _marked_overloads:
            _marked_overloads[func_params_key] = {}
        _marked_overloads[func_params_key][key] = value
        return overload(func)
    return wrapper

def get_mark_by_func(func: Callable, key: str) -> Any:
    """
    Given a callable object and a specified key, get the pre-set mark.
    """
    func_key = f"{func.__module__}.{func.__qualname__}"
    params_key = hash_kw_default_params(func)
    func_params_key = f"{func_key}::{params_key}"
    return _marked_overloads[func_params_key][key]

def get_param_names_by_kind(
        func: Callable, 
        param_kind: enum.IntEnum,
        exclude_default: bool = False,
    ) -> List[str]:
    """
    Get the names of parameters of a function by the kind of the parameter.

    Parameters
    ----------
    func : Callable
        The function to get the parameter names from.
    param_kind : enum.IntEnum
        The kind of the parameter. One of five possible values:
        - inspect.Parameter.POSITIONAL_ONLY
        - inspect.Parameter.POSITIONAL_OR_KEYWORD
        - inspect.Parameter.VAR_POSITIONAL
        - inspect.Parameter.KEYWORD_ONLY
        - inspect.Parameter.VAR_KEYWORD
    exclude_default : bool
        Whether to exclude the default parameters.

    Returns
    -------
    List[str]
        A list of parameter names.
    """
    sig = inspect.signature(func)
    param_names = []
    for name, param in sig.parameters.items():
        if param.kind == param_kind:
            if exclude_default and param.default is not inspect.Parameter.empty:
                continue
            param_names.append(name)
    return param_names

def get_param_names_all_kinds(
        func: Callable, 
        exclude_default: bool = False,
    ) -> Dict[enum.IntEnum, List[Tuple[str, Any]]]:
    """
    Get the names of parameters of a function.

    Parameters
    ----------
    func : Callable
        The function to get the parameter names from.
    exclude_default : bool
        Whether to exclude the default parameters.

    Returns
    -------
    Dict[enum.IntEnum, List[Tuple[str, Any]]]
        A dictionary of parameter names by the kind of the parameter.
        The key is the kind of the parameter, which is one of five possible values:
        - inspect.Parameter.POSITIONAL_ONLY
        - inspect.Parameter.POSITIONAL_OR_KEYWORD
        - inspect.Parameter.VAR_POSITIONAL
        - inspect.Parameter.KEYWORD_ONLY
        - inspect.Parameter.VAR_KEYWORD
    """
    sig = inspect.signature(func)
    param_names_dict = {}
    for name, param in sig.parameters.items():
        if exclude_default and param.default is not inspect.Parameter.empty:
            continue
        if param.kind not in param_names_dict:
            param_names_dict[param.kind] = []
        
        if param.default is inspect.Parameter.empty:
            param_names_dict[param.kind].append((name, inspect._empty))
        else:
            param_names_dict[param.kind].append((name, param.default))
    return param_names_dict

def hash_kw_default_params(func: Callable) -> int:
    hashable = tuple(sorted((k, v) for k, v in func.__kwdefaults__.items()))
    return deep_hash(hashable)

def list_default_params_of_each_overload(func: Callable) -> List[Dict[str, Any]]:
    """
    Returns a list of dictionaries, each containing the default parameter values 
    for one overload of the given function.
    """
    overloaded_funcs = get_overloads(func)
    params_defaults_list = []
    for ov_func in overloaded_funcs:
        params_defaults_list.append(ov_func.__kwdefaults__)
    return params_defaults_list

def load_qualified_class_or_func(full_qualified_name: str):
    parts = full_qualified_name.split('.')
    if len(parts) < 2:
        raise ValueError(f"Invalid qualified name: '{full_qualified_name}'. Two parts needed at least.")
    
    # Try importing the module step by step until it succeeds
    for i in range(len(parts) - 1, 0, -1):
        module_path = '.'.join(parts[:i])
        try:
            module = importlib.import_module(module_path)
            break
        except ImportError:
            continue
    else:
        raise ModuleNotFoundError(f"Import module failed from path: '{full_qualified_name}'")
    
    # The remaining parts are the qualified name of the class
    cls_path_parts = parts[i:]
    
    # Use getattr step by step to access nested classes
    obj = module
    try:
        for attr in cls_path_parts:
            obj = getattr(obj, attr)
    except AttributeError as e:
        raise ImportError(f"Class not found in path: '{full_qualified_name}' due to error: {e}")

    return obj



from typing import List, Union
from pydantic import BaseModel, Field

from bridgic.core.model.types import Message, Role

class BasePromptTemplate(BaseModel):
    """
    Abstract base class for prompt templates.
    
    This class provides a common interface for messages from template strings with variable substitutions.    
    
    Attributes
    ----------
    template_str : str
        The template string containing placeholders for variable substitution.
        The specific placeholder syntax depends on the concrete implementation
        (e.g., f-string, Jinja2, etc.).
    
    Methods
    -------
    format_message(role, **kwargs)
        Format a single message from the template.
    format_messages(**kwargs)
        Format multiple messages from the template.
    
    Notes
    -----
    This is an abstract base class that must be subclassed to provide
    concrete implementations. Subclasses should implement the `format_message`
    and `format_messages` methods according to their specific template
    formatting requirements.
    
    Examples
    --------
    >>> class MyTemplate(BasePromptTemplate):
    ...     def format_message(self, role=Role.USER, **kwargs):
    ...         # Implementation here
    ...         pass
    ...     
    ...     def format_messages(self, **kwargs):
    ...         # Implementation here
    ...         pass
    >>> 
    >>> template = MyTemplate(template_str="Hello {name}!")
    >>> message = template.format_message(name="World")
    """

    template_str: str

    def format_message(self, role: Union[Role, str] = Role.USER, **kwargs) -> Message:
        """
        Format a single message from the template.
        
        Parameters
        ----------
        role : Union[Role, str], default=Role.USER
            The role of the message (e.g., 'user', 'assistant', 'system').
        **kwargs
            Additional keyword arguments to be substituted into the template.
            
        Returns
        -------
        Message
            A formatted message object.
            
        Raises
        ------
        NotImplementedError
            This method must be implemented by subclasses.
        """
        raise NotImplementedError(f"format_message is not implemented in class {self.__class__.__name__}")

    def format_messages(self, **kwargs) -> List[Message]:
        """
        Format multiple messages from the template.
        
        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be substituted into the template.
            
        Returns
        -------
        List[Message]
            A list of formatted message objects.
            
        Raises
        ------
        NotImplementedError
            This method must be implemented by subclasses.
        """
        raise NotImplementedError(f"format_messages is not implemented in class {self.__class__.__name__}")


"""
The Model Protocols module defines high-level interface protocols for model interaction.

This module contains several important interface protocol definitions to provide 
capabilities needed in real-world application development, such as tool selection and 
structured output. These interfaces have clear input and output definitions and are 
"model-neutral", aiming to reduce the details developers need to consider when 
implementing features, thereby improving development efficiency.
"""

from bridgic.core.model.protocols._tool_selection import ToolSelection
from bridgic.core.model.protocols._structured_output import (
    Constraint,
    PydanticModel,
    JsonSchema,
    EbnfGrammar,
    LarkGrammar,
    Regex,
    Choice,
    RegexPattern,
    StructuredOutput,
)

__all__ = [
    "ToolSelection",
    "StructuredOutput",
    "Constraint",
    "PydanticModel",
    "JsonSchema",
    "EbnfGrammar",
    "LarkGrammar",
    "Regex",
    "RegexPattern",
    "Choice",
]


from typing import Tuple, Dict, Any, List
from inspect import Parameter, _ParameterKind


def safely_map_args(
    in_args: Tuple[Any, ...], 
    in_kwargs: Dict[str, Any],
    rx_param_names_dict: Dict[_ParameterKind, List[Tuple[str, Any]]],
) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """
    Safely map input arguments to match a target function's parameter signature.

    This function filters and adjusts positional and keyword arguments to ensure they
    can be safely passed to a target function. It handles conflicts between positional
    and keyword arguments, and filters out invalid keyword arguments when the target
    function doesn't accept **kwargs.

    Parameters
    ----------
    in_args : Tuple[Any, ...]
        Input positional arguments to be mapped.
    in_kwargs : Dict[str, Any]
        Input keyword arguments to be mapped.
    rx_param_names_dict : Dict[_ParameterKind, List[Tuple[str, Any]]]
        Target function's parameter signature dictionary, organized by parameter kind
        (POSITIONAL_ONLY, POSITIONAL_OR_KEYWORD, KEYWORD_ONLY, VAR_POSITIONAL, VAR_KEYWORD).

    Returns
    -------
    Tuple[Tuple[Any, ...], Dict[str, Any]]
        Mapped positional and keyword arguments that can be safely passed to the target function.
    """
        
    def get_param_names(param_names_dict: List[Tuple[str, Any]]) -> List[str]:
        return [name for name, _ in param_names_dict]
        
    positional_only_param_names = get_param_names(rx_param_names_dict.get(Parameter.POSITIONAL_ONLY, []))
    positional_or_keyword_param_names = get_param_names(rx_param_names_dict.get(Parameter.POSITIONAL_OR_KEYWORD, []))

    # Resolve the positional arguments `rx_args`.
    positional_param_names = positional_only_param_names + positional_or_keyword_param_names
    var_positional_param_names = get_param_names(rx_param_names_dict.get(Parameter.VAR_POSITIONAL, []))
    if len(in_args) == 1 and in_args[0] is None and len(positional_param_names) == 0 and len(var_positional_param_names) == 0:
        # The special case of the predecessor returning None and the successor has no arguments expected.
        rx_args = ()
    else:
        # In normal cases, positional arguments are unchanged.
        rx_args = in_args

    # Resolve the keyword arguments `rx_kwargs`.
    # keyword arguments are firstly filtered by `masked_keyword_param_names`.
    if len(in_args) > len(positional_only_param_names):
        masked_keyword_param_names = positional_or_keyword_param_names[:len(in_args)-len(positional_only_param_names)]
    else:
        masked_keyword_param_names = []
    rx_kwargs = {k:v for k,v in in_kwargs.items() if k not in masked_keyword_param_names}
    var_keyword_param_names = get_param_names(rx_param_names_dict.get(Parameter.VAR_KEYWORD, []))
    if not var_keyword_param_names:
        # keyword arguments are secondly filtered by `keyword_param_names`.
        keyword_only_param_names = get_param_names(rx_param_names_dict.get(Parameter.KEYWORD_ONLY, []))
        keyword_param_names = positional_or_keyword_param_names + keyword_only_param_names
        rx_kwargs = {k:v for k,v in rx_kwargs.items() if k in keyword_param_names}
    return rx_args, rx_kwargs


import json

from typing import Optional, Dict, Any, List, Union, Callable, cast, Tuple
from typing_extensions import override
from concurrent.futures import ThreadPoolExecutor
from jinja2 import Environment, PackageLoader, Template

from bridgic.core.automa import Automa, GraphAutoma, worker, RunningOptions
from bridgic.core.automa.args import From, ArgsMappingRule, System
from bridgic.core.model.protocols import ToolSelection
from bridgic.core.model.types import Tool, ToolCall
from bridgic.core.automa.interaction import InteractionFeedback
from bridgic.core.agentic.types._chat_message import ChatMessage, SystemMessage, UserTextMessage, AssistantTextMessage, ToolMessage
from bridgic.core.agentic.types._chat_message import FunctionToolCall, Function
from bridgic.core.agentic.tool_specs import ToolSpec, FunctionToolSpec, AutomaToolSpec
from bridgic.core.agentic.workers import ToolSelectionWorker

DEFAULT_MAX_ITERATIONS = 20
DEFAULT_TEMPLATE_FILE = "tools_chat.jinja"

class ReActAutoma(GraphAutoma):
    """
    A react automa is a subclass of graph automa that implements the [ReAct](https://arxiv.org/abs/2210.03629) prompting framework.
    """

    _llm: ToolSelection
    """ The LLM to be used by the react automa. """
    _tools: Optional[List[ToolSpec]]
    """ The candidate tools to be used by the react automa. """
    _system_prompt: Optional[SystemMessage]
    """ The system prompt to be used by the react automa. """
    _max_iterations: int
    """ The maximum number of iterations for the react automa. """
    _prompt_template: str
    """ The template file for the react automa. """
    _jinja_env: Environment
    """ The Jinja environment to be used by the react automa. """
    _jinja_template: Template
    """ The Jinja template to be used by the react automa. """

    def __init__(
        self,
        llm: ToolSelection,
        system_prompt: Optional[Union[str, SystemMessage]] = None,
        tools: Optional[List[Union[Callable, Automa, ToolSpec]]] = None,
        name: Optional[str] = None,
        thread_pool: Optional[ThreadPoolExecutor] = None,
        running_options: Optional[RunningOptions] = None,
        max_iterations: int = DEFAULT_MAX_ITERATIONS,
        prompt_template: str = DEFAULT_TEMPLATE_FILE,
    ):
        super().__init__(name=name, thread_pool=thread_pool, running_options=running_options)

        self._llm = llm
        if system_prompt:
            # Validate SystemMessage...
            if isinstance(system_prompt, str):
                system_prompt = SystemMessage(role="system", content=system_prompt)
            elif ("role" not in system_prompt) or (system_prompt["role"] != "system"):
                raise ValueError(f"Invalid `system_prompt` value received: {system_prompt}. It should contain `role`=`system`.")

        self._system_prompt = system_prompt
        if tools:
            self._tools = [self._ensure_tool_spec(tool) for tool in tools]
        else:
            self._tools = None
        self._max_iterations = max_iterations
        self._prompt_template = prompt_template
        self._jinja_env = Environment(loader=PackageLoader("bridgic.core.agentic.react"))
        self._jinja_template = self._jinja_env.get_template(prompt_template)

        self.add_worker(
            key="tool_selector",
            worker=ToolSelectionWorker(tool_selection_llm=llm),
            dependencies=["assemble_context"],
            args_mapping_rule=ArgsMappingRule.UNPACK,
        )

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()
        state_dict["tools"] = self._tools
        state_dict["system_prompt"] = self._system_prompt
        state_dict["llm"] = self._llm
        state_dict["max_iterations"] = self._max_iterations
        state_dict["prompt_template"] = self._prompt_template
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)
        self._max_iterations = state_dict["max_iterations"]
        self._prompt_template = state_dict["prompt_template"]
        self._tools = state_dict["tools"]
        self._system_prompt = state_dict["system_prompt"]
        self._llm = state_dict["llm"]
        self._jinja_env = Environment(loader=PackageLoader("bridgic.core.agentic.react"))
        self._jinja_template = self._jinja_env.get_template(self._prompt_template)

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    @max_iterations.setter
    def max_iterations(self, max_iterations: int) -> None:
        self._max_iterations = max_iterations

    @property
    def prompt_template(self) -> str:
        return self._prompt_template

    @prompt_template.setter
    def prompt_template(self, prompt_template: str) -> None:
        self._prompt_template = prompt_template

    @override
    async def arun(
        self,
        user_msg: Optional[Union[str, UserTextMessage]] = None,
        *,
        chat_history: Optional[List[Union[UserTextMessage, AssistantTextMessage, ToolMessage]]] = None,
        messages: Optional[List[ChatMessage]] = None,
        tools: Optional[List[Union[Callable, Automa, ToolSpec]]] = None,
        feedback_data: Optional[Union[InteractionFeedback, List[InteractionFeedback]]] = None,
    ) -> Any:
        return await super().arun(
            user_msg=user_msg,
            chat_history=chat_history,
            messages=messages,
            tools=tools,
            feedback_data=feedback_data,
        )

    @worker(is_start=True)
    async def validate_and_transform(
        self,
        user_msg: Optional[Union[str, UserTextMessage]] = None,
        *,
        chat_history: Optional[List[Union[UserTextMessage, AssistantTextMessage, ToolMessage]]] = None,
        messages: Optional[List[ChatMessage]] = None,
        tools: Optional[List[Union[Callable, Automa, ToolSpec]]] = None,
    ) -> Dict[str, Any]:
        """
        Validate and transform the input messages and tools to the canonical format.
        """
        
        # Part One: validate and transform the input messages.
        # Unify input messages of various types to the `ChatMessage` format.
        chat_messages: List[ChatMessage] = []
        if messages:
            # If `messages` is provided, use it directly.
            chat_messages = messages
        elif user_msg:
            # Since `messages` is not provided, join the system prompt + `chat_history` + `user_msg`
            # First, append the `system_prompt`
            if self._system_prompt:
                chat_messages.append(self._system_prompt)
            
            # Second, append the `chat_history`
            if chat_history:
                for history_msg in chat_history:
                    # Validate the history messages...
                    role = history_msg["role"]
                    if role == "user" or role == "assistant" or role == "tool":
                        chat_messages.append(history_msg)
                    else:
                        raise ValueError(f"Invalid role: `{role}` received in history message: `{history_msg}`, expected `user`, `assistant`, or `tool`.")
            
            # Third, append the `user_msg`
            if isinstance(user_msg, str):
                chat_messages.append(UserTextMessage(role="user", content=user_msg))
            elif isinstance(user_msg, dict):
                if "role" in user_msg and user_msg["role"] == "user":
                    chat_messages.append(user_msg)
                else:
                    raise ValueError(f"`role` must be `user` in user message: `{user_msg}`.")
        else:
            raise ValueError(f"Either `messages` or `user_msg` must be provided.")

        # Part Two: validate and transform the intput tools.
        # Unify input tools of various types to the `ToolSpec` format.
        if self._tools:
            tool_spec_list = self._tools
        elif tools:
            tool_spec_list = [self._ensure_tool_spec(tool) for tool in tools]
        else:
            # TODO: whether to support empty tool list?
            tool_spec_list = []
    
        return {
            "initial_messages": chat_messages,
            "candidate_tools": tool_spec_list,
        }

    @worker(dependencies=["validate_and_transform"], args_mapping_rule=ArgsMappingRule.UNPACK)
    async def assemble_context(
        self,
        *,
        initial_messages: Optional[List[ChatMessage]] = None,
        candidate_tools: Optional[List[ToolSpec]] = None,
        tool_selection_outputs: Tuple[List[ToolCall], Optional[str]] = From("tool_selector", default=None),
        tool_result_messages: Optional[List[ToolMessage]] = None,
        rtx = System("runtime_context"),
    ) -> Dict[str, Any]:
        # print(f"\n******* ReActAutoma.assemble_context *******\n")
        # print(f"initial_messages: {initial_messages}")
        # print(f"candidate_tools: {candidate_tools}")
        # print(f"tool_selection_outputs: {tool_selection_outputs}")
        # print(f"tool_result_messages: {tool_result_messages}")
    
        local_space = self.get_local_space(rtx)
        # Build messages memory with help of local space.
        messages_memory: List[ChatMessage] = []
        if initial_messages:
            # If `messages` is provided, use it to re-initialize the messages memory.
            messages_memory = initial_messages.copy()
        else:
            messages_memory = local_space.get("messages_memory", [])
        if tool_selection_outputs:
            # Transform tools_calls format:
            tool_calls = tool_selection_outputs[0]
            tool_calls_list = [
                FunctionToolCall(
                    id=tool_call.id,
                    type="function",
                    function=Function(
                        name=tool_call.name,
                        arguments=tool_call.arguments,
                    ),
                ) for tool_call in tool_calls
            ]
            llm_response = tool_selection_outputs[1]
            assistant_message = AssistantTextMessage(
                role="assistant",
                # TOD: name?
                content=llm_response,
                tool_calls=tool_calls_list,
            )
            messages_memory.append(assistant_message)
        if tool_result_messages:
            messages_memory.extend(tool_result_messages)
        local_space["messages_memory"] = messages_memory
        # print("--------------------------------")
        # print(f"messages_memory: {messages_memory}")
        
        # Save & retrieve tools with help of local space.
        if candidate_tools:
            local_space["tools"] = candidate_tools
        else:
            candidate_tools = local_space.get("tools", [])

        # Note: here 'messages' and `tools` are injected into the template as variables.
        raw_prompt = self._jinja_template.render(messages=messages_memory, tools=candidate_tools)
        # print(f"\n ##### raw_prompt ##### \n{raw_prompt}")

        # Note: the jinjia template must conform to the TypedDict `ChatMessage` format (in json).
        llm_messages = cast(List[ChatMessage], json.loads(raw_prompt))
        llm_tools: List[Tool] = [tool.to_tool() for tool in candidate_tools]
        
        return {
            "messages": llm_messages,
            "tools": llm_tools,
        }

    @worker(dependencies=["tool_selector"], args_mapping_rule=ArgsMappingRule.UNPACK)
    async def plan_next_step(
        self,
        tool_calls: List[ToolCall],
        llm_response: Optional[str] = None,
        messages_and_tools: dict = From("validate_and_transform"),
        rtx = System("runtime_context"),
    ) -> None:
        local_space = self.get_local_space(rtx)
        iterations_count = local_space.get("iterations_count", 0)
        iterations_count += 1
        local_space["iterations_count"] = iterations_count
        if iterations_count > self._max_iterations:
            # TODO: how to report this to users?
            self.ferry_to(
                "finally_summarize", 
                final_answer=f"Sorry, I am unable to answer your question after {self._max_iterations} iterations. Please try again later."
            )
            return

        # TODO: maybe hand over the control flow to users?
        # print(f"\n******* ReActAutoma.plan_next_step *******\n")
        # print(f"tool_calls: {tool_calls}")
        # print(f"llm_response: {llm_response}")
        if tool_calls:
            tool_spec_list = messages_and_tools["candidate_tools"]
            matched_list = self._match_tool_calls_and_tool_specs(tool_calls, tool_spec_list)
            if matched_list:
                matched_tool_calls = []
                tool_worker_keys = []
                for tool_call, tool_spec in matched_list:
                    matched_tool_calls.append(tool_call)
                    tool_worker = tool_spec.create_worker()
                    worker_key = f"tool_{tool_call.name}_{tool_call.id}"
                    self.add_worker(
                        key=worker_key,
                        worker=tool_worker,
                    )
                    # TODO: convert tool_call.arguments to the tool parameters types
                    # TODO: validate the arguments against the tool parameters / json schema
                    self.ferry_to(worker_key, **tool_call.arguments)
                    tool_worker_keys.append(worker_key)
                self.add_func_as_worker(
                    key="merge_tools_results",
                    func=self.merge_tools_results,
                    dependencies=tool_worker_keys,
                    args_mapping_rule=ArgsMappingRule.MERGE,
                )
                return matched_tool_calls
            else:
                # TODO
                ...
        else:
            # Got final answer from the LLM.
            self.ferry_to("finally_summarize", final_answer=llm_response)

    async def merge_tools_results(
        self, 
        tool_results: List[Any],
        tool_calls: List[ToolCall] = From("plan_next_step"),
    ) -> List[ToolMessage]:
        """
        Merge the results of the tools.
        """
        # print(f"\n******* ReActAutoma.merge_tools_results *******\n")
        # print(f"tool_results: {tool_results}")
        # print(f"tool_calls: {tool_calls}")
        assert len(tool_results) == len(tool_calls)
        tool_messages = []
        for tool_result, tool_call in zip(tool_results, tool_calls):
            tool_messages.append(ToolMessage(
                role="tool", 
                # Note: Convert the tool result to string, since a tool can return any type of data.
                # TODO: maybe we can use a better way to serialize the tool result?
                content=str(tool_result), 
                tool_call_id=tool_call.id
            ))
            # Remove the tool workers
            self.remove_worker(f"tool_{tool_call.name}_{tool_call.id}")
        # Remove self...
        self.remove_worker("merge_tools_results")
        self.ferry_to("assemble_context", tool_result_messages=tool_messages)
        return tool_messages

    @worker(is_output=True)
    async def finally_summarize(self, final_answer: str) -> str:
        return final_answer

    def _ensure_tool_spec(self, tool: Union[Callable, Automa, ToolSpec]) -> ToolSpec:
        if isinstance(tool, ToolSpec):
            return tool
        elif isinstance(tool, type) and issubclass(tool, Automa):
            return AutomaToolSpec.from_raw(tool)
        elif isinstance(tool, Callable):
            # Note: this test against `Callable` should be placed at last.
            return FunctionToolSpec.from_raw(tool)
        else:
            raise TypeError(f"Invalid tool type: {type(tool)} detected, expected `Callable`, `Automa`, or `ToolSpec`.")

    def _match_tool_calls_and_tool_specs(
        self,
        tool_calls: List[ToolCall],
        tool_spec_list: List[ToolSpec],
    ) -> List[Tuple[ToolCall, ToolSpec]]:
        """
        This function is used to match the tool calls and the tool specs based on the tool name.

        Parameters
        ----------
        tool_calls : List[ToolCall]
            The tool calls to match.
        tool_spec_list : List[ToolSpec]
            The tool specs to match.

        Returns
        -------
        List[(ToolCall, ToolSpec)]
            The matched tool calls and tool specs.
        """
        matched_list: List[Tuple[ToolCall, ToolSpec]] = []
        for tool_call in tool_calls:
            for tool_spec in tool_spec_list:
                if tool_call.name == tool_spec.tool_name:
                    matched_list.append((tool_call, tool_spec))
        return matched_list


from pydantic import BaseModel
from typing import Protocol, runtime_checkable, Dict, Any
from abc import abstractmethod

@runtime_checkable
class Serializable(Protocol):
    """
    Serializable is a protocol that defines the interfaces that customizes serialization.
    """
    @abstractmethod
    def dump_to_dict(self) -> Dict[str, Any]:
        """
        Dump the object to a dictionary, which will finally be serialized to bytes.
        """
        ...

    @abstractmethod
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Load the object state from a dictionary previously obtained by deserializing from bytes.
        """
        ...

@runtime_checkable
class Picklable(Protocol):
    """
    Picklable is a protocol that defines the interfaces that customizes serialization using pickle.

    Notes
    -----
    If a class implements both Serializable and Picklable, the object of the class will be 
    serialized using the implementation provided by Serializable instead of using pickle.
    """

    def __picklable_marker__(self) -> None:
        """
        This is just a marker method to distinguish Picklable objects from other objects.
        Since it is not necessary to implement this method in the subclass, thus no 
        @abstractmethod is used here.
        """
        ...



import asyncio
import inspect
import json
import threading
import uuid

from inspect import Parameter, _ParameterKind
from concurrent.futures import ThreadPoolExecutor
from typing import Any, List, Dict, Set, Mapping, Callable, Tuple, Optional, Literal, Union, ClassVar, TYPE_CHECKING
from typing_extensions import override
from types import MethodType
from dataclasses import dataclass
from pydantic import BaseModel, Field, ConfigDict

from bridgic.core.utils._console import printer
from bridgic.core.utils._msgpackx import dump_bytes
from bridgic.core.utils._args_map import safely_map_args
from bridgic.core.types._error import *
from bridgic.core.types._common import AutomaType
from bridgic.core.automa import Automa, Snapshot
from bridgic.core.automa.worker import CallableWorker, Worker
from bridgic.core.automa.interaction import Interaction, InteractionFeedback, InteractionException
from bridgic.core.automa._automa import _InteractionAndFeedback, _InteractionEventException, RunningOptions
from bridgic.core.automa.worker._worker_callback import WorkerCallback, WorkerCallbackBuilder, try_handle_error_with_callbacks
from bridgic.core.config import GlobalSetting
from bridgic.core.automa._graph_meta import GraphMeta
from bridgic.core.automa.args._args_binding import ArgsManager, ArgsMappingRule, ResultDispatchingRule

class _GraphAdaptedWorker(Worker):
    """
    A decorated worker used for GraphAutoma orchestration and scheduling. Not intended for external use.
    Follows the `Decorator` design pattern:
    - https://web.archive.org/web/20031204182047/http://patterndigest.com/patterns/Decorator.html
    - https://en.wikipedia.org/wiki/Decorator_pattern
    New Behavior Added: In addition to the original Worker functionality, maintains configuration and state variables related to dynamic scheduling and graph topology.
    """
    key: str
    dependencies: List[str]
    is_start: bool
    is_output: bool
    args_mapping_rule: str
    result_dispatching_rule: str
    _decorated_worker: Worker
    _worker_callbacks: List[WorkerCallback]

    def __init__(
        self,
        *,
        key: Optional[str] = None,
        worker: Optional[Worker] = None,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ):
        super().__init__()
        self.key = key or f"autokey-{uuid.uuid4().hex[:8]}"
        self.dependencies = dependencies
        self.is_start = is_start
        self.is_output = is_output
        self.args_mapping_rule = args_mapping_rule
        self.result_dispatching_rule = result_dispatching_rule
        self._decorated_worker = worker
        self._worker_callbacks = [cb.build() for cb in callback_builders]

    @override
    def get_report_info(self) -> Dict[str, Any]:
        report_info = super().get_report_info()
        report_info["key"] = self.key
        report_info["dependencies"] = self.dependencies
        report_info["is_start"] = self.is_start
        report_info["is_output"] = self.is_output
        report_info["args_mapping_rule"] = self.args_mapping_rule
        report_info["result_dispatching_rule"] = self.result_dispatching_rule
        return report_info
    
    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()
        state_dict["key"] = self.key
        state_dict["dependencies"] = self.dependencies
        state_dict["is_start"] = self.is_start
        state_dict["is_output"] = self.is_output
        state_dict["args_mapping_rule"] = self.args_mapping_rule
        state_dict["result_dispatching_rule"] = self.result_dispatching_rule
        state_dict["decorated_worker"] = self._decorated_worker
        state_dict["worker_callbacks"] = self._worker_callbacks
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)
        self.key = state_dict["key"]
        self.dependencies = state_dict["dependencies"]
        self.is_start = state_dict["is_start"]
        self.is_output = state_dict["is_output"]
        self.args_mapping_rule = state_dict["args_mapping_rule"]
        self.result_dispatching_rule = state_dict["result_dispatching_rule"]
        self._decorated_worker = state_dict["decorated_worker"]
        self._worker_callbacks = state_dict["worker_callbacks"]
    #
    # Delegate all the properties and methods of _GraphAdaptedWorker to the decorated worker.
    # TODO: Maybe 'Worker' should be a Protocol.
    #
    @override
    async def arun(self, *args, **kwargs) -> Any:
        for callback in self._worker_callbacks:
            await callback.on_worker_start(
                key=self.key,
                is_top_level=False,
                parent=self.parent,
                arguments={"args": args, "kwargs": kwargs},
            )

        try:
            result = await self._decorated_worker.arun(*args, **kwargs)
        except Exception as e:
            # Try to handle the exception with callbacks
            handled = await try_handle_error_with_callbacks(
                callbacks=self._worker_callbacks,
                key=self.key,
                is_top_level=False,
                parent=self.parent,
                arguments={"args": args, "kwargs": kwargs},
                error=e,
            )

            # If no callback handled the exception, re-raise it
            if not handled:
                raise

            # If exception was handled, set result to None
            result = None

        for callback in self._worker_callbacks:
            await callback.on_worker_end(
                key=self.key,
                is_top_level=False,
                parent=self.parent,
                arguments={"args": args, "kwargs": kwargs},
                result=result,
            )
        return result

    @override
    def get_input_param_names(self) -> Dict[_ParameterKind, List[Tuple[str, Any]]]:
        return self._decorated_worker.get_input_param_names()

    @property
    def parent(self) -> "Automa":
        return self._decorated_worker.parent

    @parent.setter
    def parent(self, value: "Automa"):
        self._decorated_worker.parent = value

    @override
    def __str__(self) -> str:
        # TODO: need some refactoring
        return str(self._decorated_worker)

    @override
    def __eq__(self, other):
        if self is other:
            return True
        return self._decorated_worker == other

    def is_automa(self) -> bool:
        return isinstance(self._decorated_worker, Automa)

    def get_decorated_worker(self) -> Worker:
        return self._decorated_worker

@dataclass
class _RunnningTask:
    """
    States of the current running task.
    The instances of this class do not need to be serialized.
    """
    worker_key: str
    task: asyncio.Task

class _AutomaInputBuffer(BaseModel):
    args: Tuple[Any, ...] = ()
    kwargs: Dict[str, Any] = {}

class _KickoffInfo(BaseModel):
    # The key of the worker that is going to be kicked off.
    worker_key: str
    # Worker key or the container "__automa__"
    last_kickoff: Optional[str]
    # Whether the kickoff is triggered by ferry_to() initiated by developers.
    from_ferry: bool = False
    # Whether the run is finished.
    # Finished workers may be kicked off again after a human interaction and thus should be skipped.
    run_finished: bool = False
    args: Tuple[Any, ...] = ()
    kwargs: Dict[str, Any] = {}

class _WorkerDynamicState(BaseModel):
    # Dynamically record the dependency workers keys of each worker.
    # Will be reset to the topology edges/dependencies of the worker
    # once the task is finished or the topology is changed.
    dependency_triggers: Set[str]

class _AddWorkerDeferredTask(BaseModel):
    task_type: Literal["add_worker"] = Field(default="add_worker")
    worker_key: str
    worker_obj: Worker # Note: Not a pydantic model!!
    dependencies: List[str] = []
    is_start: bool = False
    is_output: bool = False
    args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS
    result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS
    callback_builders: List[WorkerCallbackBuilder] = []

    model_config = ConfigDict(arbitrary_types_allowed=True)

class _RemoveWorkerDeferredTask(BaseModel):
    task_type: Literal["remove_worker"] = Field(default="remove_worker")
    worker_key: str

class _AddDependencyDeferredTask(BaseModel):
    task_type: Literal["add_dependency"] = Field(default="add_dependency")
    worker_key: str
    dependency: str

class _FerryDeferredTask(BaseModel):
    ferry_to_worker_key: str
    kickoff_worker_key: Optional[str]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]

class GraphAutoma(Automa, metaclass=GraphMeta):
    """
    Dynamic Directed Graph (abbreviated as DDG) implementation of Automa. `GraphAutoma` manages 
    the running control flow between workers automatically, via `dependencies` and `ferry_to`.
    Outputs of workers can be mapped and passed to their successor workers in the runtime, 
    following `args_mapping_rule` and `result_dispatching_rule`.

    Parameters
    ----------
    name : Optional[str]
        The name of the automa.

    thread_pool : Optional[ThreadPoolExecutor]
        The thread pool for parallel running of I/O-bound tasks.

        - If not provided, a default thread pool will be used.
        The maximum number of threads in the default thread pool dependends on the number of CPU cores. Please refer to 
        the [ThreadPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor) for detail.

        - If provided, all workers (including all nested Automa instances) will be run in it. In this case, the 
        application layer code is responsible to create it and shut it down.

    Examples
    --------

    The following example shows how to use `GraphAutoma` to create a simple graph automa that prints "Hello, Bridgic".

    ```python
    import asyncio
    from bridgic.core.automa import GraphAutoma, worker, ArgsMappingRule

    class MyGraphAutoma(GraphAutoma):
        @worker(is_start=True)
        async def greet(self) -> list[str]:
            return ["Hello", "Bridgic"]

        @worker(dependencies=["greet"], args_mapping_rule=ArgsMappingRule.AS_IS, result_dispatching_rule=ResultDispatchingRule.AS_IS, is_output=True)
        async def output(self, message: list[str]):
            print("Echo: " + " ".join(message))

    async def main():
        automa_obj = MyGraphAutoma(name="my_graph_automa")
        await automa_obj.arun()

    asyncio.run(main())
    ```
    """

    # Automa type.
    AUTOMA_TYPE: ClassVar[AutomaType] = AutomaType.Graph

    # The initial topology defined by @worker functions.
    _registered_worker_funcs: ClassVar[Dict[str, Callable]] = {}

    # IMPORTANT: The entire states of a GraphAutoma instance include 2 part:
    # 
    # Part-1 (for the states of topology structure):
    #   1. Inner worker instances: self._workers
    #   2. Relations between worker: self._worker_forwards
    #   3. Dynamic states that serve as trigger of execution of workers: self._workers_dynamic_states
    #   4. Execution result of inner workers: self._worker_output
    #   5. Configurations of this automa instance: self._output_worker_key
    # 
    # Part-2 (for the states of running states):
    #   1. Records of Workers that are going to be kicked off: self._current_kickoff_workers
    #   2. Records of running or deferred tasks:
    #      - self._running_tasks
    #      - self._topology_change_deferred_tasks
    #      - self._ferry_deferred_tasks
    #      - self._set_output_worker_deferred_task
    #   3. Buffers of automa inputs: self._input_buffer
    #   4. Ongoing human interactions: self._ongoing_interactions
    #   ...

    _workers: Dict[str, _GraphAdaptedWorker]
    _worker_output: Dict[str, Any]
    _worker_forwards: Dict[str, List[str]]

    _current_kickoff_workers: List[_KickoffInfo]
    _input_buffer: _AutomaInputBuffer
    _workers_dynamic_states: Dict[str, _WorkerDynamicState]

    # The whole running process of the DDG is divided into two main phases:
    # 1. [Initialization Phase] The first phase (when _automa_running is False): the initial topology of DDG was constructed.
    # 2. [Running Phase] The second phase (when _automa_running is True): the DDG is running, and the workers are executed in a dynamic step-by-step manner (DS loop).
    _automa_running: bool

    #########################################################
    #### The following fields need not to be serialized. ####
    #########################################################
    _running_tasks: List[_RunnningTask]

    # TODO: The following deferred task structures need to be thread-safe.
    # TODO: Need to be refactored when parallelization features are added.
    _topology_change_deferred_tasks: List[Union[_AddWorkerDeferredTask, _RemoveWorkerDeferredTask]]
    _ferry_deferred_tasks: List[_FerryDeferredTask]

    def __init__(
        self,
        name: Optional[str] = None,
        thread_pool: Optional[ThreadPoolExecutor] = None,
        running_options: Optional[RunningOptions] = None,
    ):
        """
        Parameters
        ----------
        name : Optional[str]
            The name of the automa.

        thread_pool : Optional[ThreadPoolExecutor]
            The thread pool for parallel running of I/O-bound tasks.

            - If not provided, a default thread pool will be used.
            The maximum number of threads in the default thread pool dependends on the number of CPU cores. Please refer to 
            the [ThreadPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor) for detail.

            - If provided, all workers (including all nested Automa instances) will be run in it. In this case, the 
            application layer code is responsible to create it and shut it down.

        running_options : Optional[RunningOptions]
            Running options for this Automa instance, including callback_builders.
            If None, uses default RunningOptions.

        state_dict : Optional[Dict[str, Any]]
            A dictionary for initializing the automa's runtime states. This parameter is designed for framework use only.
        """
        super().__init__(name=name, thread_pool=thread_pool, running_options=running_options)

        self._workers = {}
        self._worker_outputs = {}
        self._automa_running = False

        # Initialize the states that need to be serialized.
        self._normal_init()

        # The list of the tasks that are currently being executed.
        self._running_tasks = []
        # deferred tasks
        self._topology_change_deferred_tasks = []
        self._ferry_deferred_tasks = []

    def _normal_init(self):
        ###############################################################################
        # Initialization of [Part One: Topology-Related Runtime States] #### Strat ####
        ###############################################################################

        cls = type(self)

        # _workers, _worker_forwards and _workers_dynamic_states will be initialized incrementally by add_worker()...
        self._worker_forwards = {}
        self._worker_output = {}
        self._workers_dynamic_states = {}

        if cls.AUTOMA_TYPE == AutomaType.Graph:
            # The _registered_worker_funcs data are from @worker decorators.
            for worker_key, worker_func in cls._registered_worker_funcs.items():
                # The decorator based mechanism (i.e. @worker) is based on the add_worker() interface.
                # Parameters check and other implementation details can be unified.
                self._add_func_as_worker_internal(
                    key=worker_key,
                    func=worker_func,
                    dependencies=worker_func.__dependencies__,
                    is_start=worker_func.__is_start__,
                    is_output=worker_func.__is_output__,
                    args_mapping_rule=worker_func.__args_mapping_rule__,
                    result_dispatching_rule=worker_func.__result_dispatching_rule__,
                    callback_builders=worker_func.__callback_builders__,
                )

        ###############################################################################
        # Initialization of [Part One: Topology-Related Runtime States] ##### End #####
        ###############################################################################

        ###############################################################################
        # Initialization of [Part Two: Task-Related Runtime States] ###### Strat ######
        ###############################################################################

        # -- Current kickoff workers list.
        # The key list of the workers that are ready to be immediately executed in the next DS (Dynamic Step). It will be lazily initialized in _compile_graph_and_detect_risks().
        self._current_kickoff_workers = []
        # -- Automa input buffer.
        self._input_buffer = _AutomaInputBuffer()

        ###############################################################################
        # Initialization of [Part Two: Task-Related Runtime States] ####### End #######
        ###############################################################################

    ###############################################################
    ########## [Bridgic Serialization Mechanism] starts ###########
    ###############################################################

    # The version of the serialization format.
    SERIALIZATION_VERSION: str = "1.0"

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()

        state_dict["name"] = self.name
        state_dict["automa_running"] = self._automa_running

        # States related to workers.
        state_dict["workers"] = self._workers
        state_dict["worker_forwards"] = self._worker_forwards
        state_dict["workers_dynamic_states"] = self._workers_dynamic_states
        state_dict["worker_output"] = self._worker_output

        # States related to interruption recovery.
        state_dict["current_kickoff_workers"] = self._current_kickoff_workers
        state_dict["input_buffer"] = self._input_buffer

        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)

        self.name = state_dict["name"]
        self._automa_running = state_dict["automa_running"]

        # States related to workers.
        self._workers = state_dict["workers"]
        for worker in self._workers.values():
            worker.parent = self
        self._worker_forwards = state_dict["worker_forwards"]
        self._workers_dynamic_states = state_dict["workers_dynamic_states"]
        self._worker_output = state_dict["worker_output"]

        # States related to interruption recovery.
        self._current_kickoff_workers = state_dict["current_kickoff_workers"]
        self._input_buffer = state_dict["input_buffer"]

        # The list of the tasks that are currently being executed.
        self._running_tasks = []
        # Deferred tasks
        self._topology_change_deferred_tasks = []
        self._set_output_worker_deferred_task = None
        self._ferry_deferred_tasks = []

    ###############################################################
    ########### [Bridgic Serialization Mechanism] ends ############
    ###############################################################

    def _add_worker_incrementally(
        self,
        key: str,
        worker: Worker,
        *,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> None:
        """
        Incrementally add a worker into the automa. For internal use only.
        This method is one of the very basic primitives of DDG for dynamic topology changes. 
        """
        if key in self._workers:
            raise AutomaRuntimeError(
                f"duplicate workers with the same key '{key}' are not allowed to be added!"
            )

        # Merge callback builders: Global -> Ancestor Automa(s) -> Current Automa -> Nested Automa (if worker is automa) -> Worker
        effective_callback_builders = []
        effective_callback_builders.extend(GlobalSetting.read().callback_builders)
        # Collect callback builders from all ancestor automas in the ancestor chain (from top-level to current)
        effective_callback_builders.extend(self._collect_ancestor_callback_builders())
        # If the worker itself is an automa, include its own RunningOptions callback builders
        if isinstance(worker, Automa):
            effective_callback_builders.extend(worker._running_options.callback_builders)
        # Include the callback builders from the worker itself.
        effective_callback_builders.extend(callback_builders)

        # Note: the dependencies argument must be a new copy of the list, created with list(dependencies).
        # Refer to the Python documentation for more details:
        # 1. https://docs.python.org/3/reference/compound_stmts.html#function-definitions
        # "Default parameter values are evaluated from left to right when the function definition is executed"
        # 2. https://docs.python.org/3/tutorial/controlflow.html#default-argument-values
        # "The default values are evaluated at the point of function definition in the defining scope"
        # "Important warning: The default value is evaluated only once."
        new_worker_obj = _GraphAdaptedWorker(
            key=key,
            worker=worker,
            dependencies=list(dependencies),
            is_start=is_start,
            is_output=is_output,
            args_mapping_rule=args_mapping_rule,
            result_dispatching_rule=result_dispatching_rule,
            callback_builders=effective_callback_builders,
        )

        # Register the worker_obj.
        new_worker_obj.parent = self
        self._workers[new_worker_obj.key] = new_worker_obj

        # Incrementally update the dynamic states of added workers.
        self._workers_dynamic_states[key] = _WorkerDynamicState(
            dependency_triggers=set(dependencies)
        )

        # Incrementally update the forwards table.
        for trigger in dependencies:
            if trigger not in self._worker_forwards:
                self._worker_forwards[trigger] = []
            self._worker_forwards[trigger].append(key)

        # If the added worker is an automa, recursively propagate callbacks to inner workers.
        if new_worker_obj.is_automa():
            nested_automa = new_worker_obj.get_decorated_worker()
            if isinstance(nested_automa, GraphAutoma):
                # Collect callback builders from all ancestor automas in the ancestor chain (from top-level to current)
                ancestor_callback_builders = self._collect_ancestor_callback_builders()
                # Append ancestor callbacks to the _cached_callbacks of the nested automa instance.
                nested_automa._cached_callbacks = nested_automa._get_automa_callbacks() + [cb.build() for cb in ancestor_callback_builders]
                # Recursively propagate ancestor callbacks to inner workers.
                self._propagate_callbacks_to_nested_automa(
                    nested_automa=nested_automa,
                    callback_builders=ancestor_callback_builders,
                )

    def _propagate_callbacks_to_nested_automa(
        self,
        nested_automa: "GraphAutoma",
        callback_builders: List[WorkerCallbackBuilder],
    ) -> None:
        """
        Recursively propagate callback builders to all workers in a nested automa.

        This method ensures that callbacks from all ancestor automas in the ancestor chain
        are applied to all workers in nested automa instances, including deeply nested ones.

        Parameters
        ----------
        nested_automa : GraphAutoma
            The nested automa instance to propagate callbacks to.
        callback_builders : List[WorkerCallbackBuilder]
            The callback builders from all ancestor automas in the ancestor chain 
            (from top-level to current) to propagate.
        """
        for worker_key in nested_automa.all_workers():
            nested_worker = nested_automa._workers[worker_key]

            # Add callback instances built from all ancestor automas' callback builders in the ancestor chain.
            new_callbacks = [cb.build() for cb in callback_builders]
            nested_worker._worker_callbacks += new_callbacks

            # Check if the nested worker is also an automa, and recursively propagate.
            if nested_worker.is_automa():
                deeper_nested_automa = nested_worker.get_decorated_worker()
                if isinstance(deeper_nested_automa, GraphAutoma):
                    # Recursively propagate to deeper nested automas.
                    # Include current nested automa's callbacks in the propagation chain,
                    # so that deeper nested workers get callbacks from all ancestor automas.
                    self._propagate_callbacks_to_nested_automa(
                        nested_automa=deeper_nested_automa,
                        callback_builders=callback_builders,
                    )

    def _remove_worker_incrementally(
        self,
        key: str
    ) -> None:
        """
        Incrementally remove a worker from the automa. For internal use only.
        This method is one of the very basic primitives of DDG for dynamic topology changes.
        """
        if key not in self._workers:
            raise AutomaRuntimeError(
                f"fail to remove worker '{key}' that does not exist!"
            )

        worker_to_remove = self._workers[key]

        # Remove the worker.
        del self._workers[key]
        # Incrementally update the dynamic states of removed workers.
        del self._workers_dynamic_states[key]

        if key in self._worker_forwards:
            # Update the dependencies of the successor workers, if needed.
            for successor in self._worker_forwards[key]:
                self._workers[successor].dependencies.remove(key)
                # Note this detail here: use discard() instead of remove() to avoid KeyError.
                # This case occurs when a worker call remove_worker() to remove its predecessor worker.
                self._workers_dynamic_states[successor].dependency_triggers.discard(key)
            # Incrementally update the forwards table.
            del self._worker_forwards[key]

        # Remove from the forwards list of all dependencies worker.
        for trigger in worker_to_remove.dependencies:
            self._worker_forwards[trigger].remove(key)
        if key in self._worker_interaction_indices:
            del self._worker_interaction_indices[key]
        if key in self._ongoing_interactions:
            del self._ongoing_interactions[key]

    def _add_dependency_incrementally(
        self,
        key: str,
        dependency: str,
    ) -> None:
        """
        Incrementally add a dependency from `key` to `depends`. For internal use only.
        This method is one of the very basic primitives of DDG for dynamic topology changes.
        """
        if key not in self._workers:
            raise AutomaRuntimeError(
                f"fail to add dependency from a worker that does not exist: `{key}`!"
            )
        if dependency not in self._workers:
            raise AutomaRuntimeError(
                f"fail to add dependency to a worker that does not exist: `{dependency}`!"
            )
        if dependency in self._workers[key].dependencies:
            raise AutomaRuntimeError(
                f"dependency from '{key}' to '{dependency}' already exists!"
            )

        self._workers[key].dependencies.append(dependency)
        # Note this detail here for dynamic states change:
        # The new dependency added here may be removed right away if the dependency is just the next kickoff worker. This is a valid behavior.
        self._workers_dynamic_states[key].dependency_triggers.add(dependency)

        if dependency not in self._worker_forwards:
            self._worker_forwards[dependency] = []
        self._worker_forwards[dependency].append(key)

    def _add_worker_internal(
        self,
        key: str,
        worker: Worker,
        *,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> None:
        """
        The private version of the method `add_worker()`.
        """

        def _basic_worker_params_check(key: str, worker_obj: Worker):
            if not isinstance(worker_obj, Worker):
                raise TypeError(
                    f"worker_obj to be registered must be a Worker, "
                    f"but got {type(worker_obj)} for worker '{key}'"
                )

            if not asyncio.iscoroutinefunction(worker_obj.arun):
                raise WorkerSignatureError(
                    f"arun of Worker must be an async method, "
                    f"but got {type(worker_obj.arun)} for worker '{key}'"
                )

            if not isinstance(dependencies, list):
                raise TypeError(
                    f"dependencies must be a list, "
                    f"but got {type(dependencies)} for worker '{key}'"
                )
            if not all([isinstance(d, str) for d in dependencies]):
                raise ValueError(
                    f"dependencies must be a List of str, "
                    f"but got {dependencies} for worker {key}"
                )

            if args_mapping_rule not in ArgsMappingRule:
                raise ValueError(
                    f"args_mapping_rule must be one of the following: {[e for e in ArgsMappingRule]}, "
                    f"but got {args_mapping_rule} for worker {key}"
                )

            if result_dispatching_rule not in ResultDispatchingRule:
                raise ValueError(
                    f"result_dispatching_rule must be one of the following: {[e for e in ResultDispatchingRule]}, "
                    f"but got {result_dispatching_rule} for worker {key}"
                )

        # Ensure the parameters are valid.
        _basic_worker_params_check(key, worker)

        if not self._automa_running:
            # Add worker during the [Initialization Phase].
            self._add_worker_incrementally(
                key=key,
                worker=worker,
                dependencies=dependencies,
                is_start=is_start,
                is_output=is_output,
                args_mapping_rule=args_mapping_rule,
                result_dispatching_rule=result_dispatching_rule,
                callback_builders=callback_builders,
            )
        else:
            # Add worker during the [Running Phase].
            deferred_task = _AddWorkerDeferredTask(
                worker_key=key,
                worker_obj=worker,
                dependencies=dependencies,
                is_start=is_start,
                is_output=is_output,
                args_mapping_rule=args_mapping_rule,
                result_dispatching_rule=result_dispatching_rule,
                callback_builders=callback_builders,
            )
            # Note1: the execution order of topology change deferred tasks is important and is determined by the order of the calls of add_worker(), remove_worker() and add_dependency() in one DS.
            # Note2: add_worker() and remove_worker() may be called in a new thread. But _topology_change_deferred_tasks is not necessary to be thread-safe due to Visibility Guarantees of the Bridgic Concurrency Model.
            self._topology_change_deferred_tasks.append(deferred_task)

    def _add_func_as_worker_internal(
        self,
        key: str,
        func: Callable,
        *,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> None:
        """
        The private version of the method `add_func_as_worker()`.
        """
        if not isinstance(func, MethodType) and key in self._registered_worker_funcs:
            func = MethodType(func, self)

        # Validate: if func is a method, its bounded __self__ must be self when add_func_as_worker() is called.
        if hasattr(func, "__self__") and func.__self__ is not self:
            raise AutomaRuntimeError(
                f"the bounded instance of `func` must be the same as the instance of the GraphAutoma, "
                f"but got {func.__self__}"
            )

        # Register func as an instance of CallableWorker.
        func_worker = CallableWorker(func)

        self._add_worker_internal(
            key=key,
            worker=func_worker,
            dependencies=dependencies,
            is_start=is_start,
            is_output=is_output,
            args_mapping_rule=args_mapping_rule,
            result_dispatching_rule=result_dispatching_rule,
            callback_builders=callback_builders,
        )

    def all_workers(self) -> List[str]:
        """
        Gets a list containing the keys of all workers registered in this Automa.

        Returns
        -------
        List[str]
            A list of worker keys.
        """
        return list(self._workers.keys())

    def add_worker(
        self,
        key: str,
        worker: Worker,
        *,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> None:
        """
        This method is used to add a worker dynamically into the automa.

        If this method is called during the [Initialization Phase], the worker will be added immediately. If this method is called during the [Running Phase], the worker will be added as a deferred task which will be executed in the next DS.

        The dependencies can be added together with a worker. However, you can add a worker without any dependencies.

        Note: args_mapping_rule and result_dispatching_rule could only be set when using worker-adding API. Even if the worker has no any dependencies.

        Parameters
        ----------
        key : str
            The key of the worker.
        worker : Worker
            The worker instance to be registered.
        dependencies : List[str]
            A list of worker keys that the worker depends on.
        is_start : bool
            Whether the worker is a start worker.
        is_output : bool
            Whether the worker is an output worker.
        args_mapping_rule : ArgsMappingRule
            The rule of arguments mapping.
        result_dispatching_rule : ResultDispatchingRule
            The rule of result dispatch.
        callback_builders : List[WorkerCallbackBuilder]
            A list of worker callback builders to be registered.
            Callback instances will be created from builders when the worker is instantiated.
        """
        self._add_worker_internal(
            key=key,
            worker=worker,
            dependencies=dependencies,
            is_start=is_start,
            is_output=is_output,
            args_mapping_rule=args_mapping_rule,
            result_dispatching_rule=result_dispatching_rule,
            callback_builders=callback_builders,
        )

    def add_func_as_worker(
        self,
        key: str,
        func: Callable,
        *,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> None:
        """
        This method is used to add a function as a worker into the automa.

        The format of the parameters will follow that of the decorator @worker(...), so that the 
        behavior of the decorated function is consistent with that of normal CallableLandableWorker objects.

        Parameters
        ----------
        key : str
            The key of the function worker.
        func : Callable
            The function to be added as a worker to the automa.
        dependencies : List[str]
            A list of worker names that the decorated callable depends on.
        is_start : bool
            Whether the decorated callable is a start worker. True means it is, while False means it is not.
        is_output : bool
            Whether the decorated callable is an output worker. True means it is, while False means it is not.
        args_mapping_rule : ArgsMappingRule
            The rule of arguments mapping.
        result_dispatching_rule : ResultDispatchingRule
            The rule of result dispatch.
        callback_builders : List[WorkerCallbackBuilder]
            A list of worker callback builders to be registered.
            Callback instances will be created from builders when the worker is instantiated.
        """
        self._add_func_as_worker_internal(
            key=key,
            func=func,
            dependencies=dependencies,
            is_start=is_start,
            is_output=is_output,
            args_mapping_rule=args_mapping_rule,
            result_dispatching_rule=result_dispatching_rule,
            callback_builders=callback_builders,
        )

    def worker(
        self,
        *,
        key: Optional[str] = None,
        dependencies: List[str] = [],
        is_start: bool = False,
        is_output: bool = False,
        args_mapping_rule: ArgsMappingRule = ArgsMappingRule.AS_IS,
        result_dispatching_rule: ResultDispatchingRule = ResultDispatchingRule.AS_IS,
        callback_builders: List[WorkerCallbackBuilder] = [],
    ) -> Callable:
        """
        This is a decorator used to mark a function as an GraphAutoma detectable Worker. Dislike the 
        global decorator @worker(...), it is usally used after an GraphAutoma instance is initialized.

        The format of the parameters will follow that of the decorator @worker(...), so that the 
        behavior of the decorated function is consistent with that of normal CallableLandableWorker objects.

        Parameters
        ----------
        key : str
            The key of the worker. If not provided, the name of the decorated callable will be used.
        dependencies : List[str]
            A list of worker names that the decorated callable depends on.
        is_start : bool
            Whether the decorated callable is a start worker. True means it is, while False means it is not.
        is_output : bool
            Whether the decorated callable is an output worker. True means it is, while False means it is not.
        args_mapping_rule : str
            The rule of arguments mapping. The options are: "auto", "as_list", "as_dict", "suppressed".
        result_dispatching_rule : ResultDispatchingRule
            The rule of result dispatch.
        callback_builders : List[WorkerCallbackBuilder]
            A list of worker callback builders to be registered.
            Callback instances will be created from builders when the worker is instantiated.
        """
        def wrapper(func: Callable):
            self._add_func_as_worker_internal(
                key=(key or func.__name__),
                func=func,
                dependencies=dependencies,
                is_start=is_start,
                is_output=is_output,
                args_mapping_rule=args_mapping_rule,
                result_dispatching_rule=result_dispatching_rule,
                callback_builders=callback_builders,
            )

        return wrapper

    def remove_worker(self, key: str) -> None:
        """
        Remove a worker from the Automa. This method can be called at any time to remove a worker from the Automa.

        When a worker is removed, all dependencies related to this worker, including all the dependencies of the worker itself and the dependencies between the worker and its successor workers, will be also removed.

        Parameters
        ----------
        key : str
            The key of the worker to be removed.

        Returns
        -------
        None

        Raises
        ------
        AutomaDeclarationError
            If the worker specified by key does not exist in the Automa, this exception will be raised.
        """
        if not self._automa_running:
            # remove immediately
            self._remove_worker_incrementally(key)
        else:
            deferred_task = _RemoveWorkerDeferredTask(
                worker_key=key,
            )
            # Note: the execution order of topology change deferred tasks is important and is determined by the order of the calls of add_worker(), remove_worker() and add_dependency() in one DS.
            self._topology_change_deferred_tasks.append(deferred_task)

    def add_dependency(
        self,
        key: str,
        dependency: str,
    ) -> None:
        """
        This method is used to dynamically add a dependency from `key` to `dependency`.

        Note: args_mapping_rule and result_dispatching_rule is not allowed to be set by this method, 
        instead they should be set together with add_worker() or add_func_as_worker() when adding the worker.

        Parameters
        ----------
        key : str
            The key of the worker that will depend on the worker with key `dependency`.
        dependency : str
            The key of the worker on which the worker with key `key` will depend.
        """
        ...
        if not self._automa_running:
            # add the dependency immediately
            self._add_dependency_incrementally(key, dependency)
        else:
            deferred_task = _AddDependencyDeferredTask(
                worker_key=key,
                dependency=dependency,
            )
            # Note: the execution order of topology change deferred tasks is important and is determined by the order of the calls of add_worker(), remove_worker() and add_dependency() in one DS.
            self._topology_change_deferred_tasks.append(deferred_task)

    def _validate_canonical_graph(self):
        """
        This method is used to validate that DDG graph is canonical.
        """
        for worker_key, worker_obj in self._workers.items():
            for dependency_key in worker_obj.dependencies:
                if dependency_key not in self._workers:
                    raise AutomaCompilationError(
                        f"the dependency `{dependency_key}` of worker `{worker_key}` does not exist"
                    )
        assert set(self._workers.keys()) == set(self._workers_dynamic_states.keys())
        for worker_key, worker_dynamic_state in self._workers_dynamic_states.items():
            for dependency_key in worker_dynamic_state.dependency_triggers:
                assert dependency_key in self._workers[worker_key].dependencies

        for worker_key, worker_obj in self._workers.items():
            for dependency_key in worker_obj.dependencies:
                assert worker_key in self._worker_forwards[dependency_key]
        for worker_key, successor_keys in self._worker_forwards.items():
            for successor_key in successor_keys:
                assert worker_key in self._workers[successor_key].dependencies

    def _compile_graph_and_detect_risks(self):
        """
        This method should be called at the very beginning of self.run() to ensure that:
        1. The whole graph is built out of all of the following worker sources:
            - Pre-defined workers, such as:
                - Methods decorated with @worker(...)
            - Post-added workers, such as:
                - Functions decorated with @automa_obj.worker(...)
                - Workers added via automa_obj.add_func_as_worker(...)
                - Workers added via automa_obj.add_worker(...)
        2. The dependencies of each worker are confirmed to satisfy the DAG constraints.
        """

        # Validate the canonical graph.
        self._validate_canonical_graph()
        # Validate the DAG constraints.
        GraphMeta.validate_dag_constraints(self._worker_forwards)
        # TODO: More validations can be added here...

        # Find all connected components of the whole automa graph.
        self._find_connected_components()

    def ferry_to(self, key: str, /, *args, **kwargs):
        """
        Defer the invocation to the specified worker, passing any provided arguments. This creates a 
        delayed call, ensuring the worker will be scheduled to run asynchronously in the next event loop, 
        independent of its dependencies.

        This primitive is commonly used for:

        1. Implementing dynamic branching based on runtime conditions.
        2. Creating logic that forms cyclic graphs.

        Parameters
        ----------
        key : str
            The key of the worker to run.
        args : optional
            Positional arguments to be passed.
        kwargs : optional
            Keyword arguments to be passed.

        Examples
        --------
        ```python
        class MyGraphAutoma(GraphAutoma):
            @worker(is_start=True)
            def start_worker(self):
                number = random.randint(0, 1)
                if number == 0:
                    self.ferry_to("cond_1_worker", number=number)
                else:
                    self.ferry_to("cond_2_worker")

            @worker()
            def cond_1_worker(self, number: int):
                print(f'Got {{number}}!')

            @worker()
            def cond_2_worker(self):
                self.ferry_to("start_worker")

        automa = MyGraphAutoma()
        await automa.arun()

        # Output: Got 0!
        ```
        """
        # TODO: check worker_key is valid, maybe deferred check...
        running_options = self._get_top_running_options()
        # if debug is enabled, trace back the kickoff worker key from stacktrace.
        kickoff_worker_key: str = self._trace_back_kickoff_worker_key_from_stack() if running_options.debug else None
        deferred_task = _FerryDeferredTask(
            ferry_to_worker_key=key,
            kickoff_worker_key=kickoff_worker_key,
            args=args,
            kwargs=kwargs,
        )
        # Note: ferry_to() may be called in a new thread.
        # But _ferry_deferred_tasks is not necessary to be thread-safe due to Visibility Guarantees of the Bridgic Concurrency Model.
        self._ferry_deferred_tasks.append(deferred_task)

    def _clean_all_worker_local_space(self):
        """
        Clean the local space of all workers.
        """
        for worker_obj in self._workers.values():
            worker_obj.local_space = {}

    async def arun(
        self,
        *args: Tuple[Any, ...],
        feedback_data: Optional[Union[InteractionFeedback, List[InteractionFeedback]]] = None,
        **kwargs: Dict[str, Any]
    ) -> Any:
        """
        The entry point for running the constructed `GraphAutoma` instance.

        This method serves as the entry point for both initial execution and resumption after 
        interruption of an automa instance. It automatically drives the execution of workers 
        based on their `dependencies` and explicit `ferry_to()` calls. Each execution will be 
        wrapped in an `asyncio.Task` to ensure context isolation.

        **Automatic Scheduling**

        The scheduling behavior in `GraphAutoma` is automatically driven by:

        - Worker dependencies: Workers are scheduled to run only after all their necessary 
          dependencies are satisfied. The dependencies automatically drive the execution order.

        - Calling ferry_to: During execution, a worker can explicitly trigger another worker 
          by calling `ferry_to()`, which enables dynamic flow control and conditional branching.

        - Dynamic topology changes: When the graph topology is modified at runtime (such as 
          adding or removing workers or dependencies), the scheduling system seamlessly updates 
          to reflect the latest structure, ensuring that worker execution always follows the 
          current graph.

        **Human Interaction Mechanism**

        Workers can request human input by calling `interact_with_human()` during execution. 
        When this occurs:

        - The execution will be paused after the running workers finish their execution.
        - The Automa's state will be serialized into a `Snapshot` object.
        - An `InteractionException` will be raised to the application layer. It contains both the 
          list of pending `Interaction` objects and the `Snapshot` object.
        - The application layer may persist the `Snapshot` properly to resume the execution later.
        - To resume execution, the application layer should reload the Automa state using 
          `load_from_snapshot()` with the saved `Snapshot` object and call `arun()` again with 
          `feedback_data` containing the user's feedback(s) to finish a complete interaction.

        Parameters
        ----------
        args : optional
            Positional arguments to be passed.
        feedback_data : Optional[Union[InteractionFeedback, List[InteractionFeedback]]]
            Feedbacks that are received from one or multiple human interactions occurred before the
            Automa was paused. This argument may be of type `InteractionFeedback` or 
            `List[InteractionFeedback]`. If only one interaction occurred, `feedback_data` should be
            of type `InteractionFeedback`. If multiple interactions occurred simultaneously, 
            `feedback_data` should be of type `List[InteractionFeedback]`.
        kwargs : optional
            Keyword arguments which may be further propagated to contained workers.

        Returns
        -------
        Any
            The execution result of the output-worker that has the setting `is_output=True`,
            otherwise None.

        Raises
        ------
        InteractionException
            If the Automa is the top-level Automa and the `interact_with_human()` method is called
            by one or more workers within the lastest event loop iteration, this exception will be
            raised to the application layer.
        """
        if self.is_top_level():
            # For top-level automa, wrap in a task to ensure context isolation
            task = asyncio.create_task(
                self._arun_internal(*args, feedback_data=feedback_data, **kwargs),
                name=f"GraphAutoma-{self.name}-arun"
            )
            return await task
        else:
            # For nested automa, directly call _arun_internal to avoid redundant task creation
            return await self._arun_internal(*args, feedback_data=feedback_data, **kwargs)

    async def _arun_internal(
        self,
        *args: Tuple[Any, ...],
        feedback_data: Optional[Union[InteractionFeedback, List[InteractionFeedback]]] = None,
        **kwargs: Dict[str, Any]
    ) -> Any:
        """
        Internal implementation of `arun()` for `GraphAutoma`.

        The scheduling behavior in `GraphAutoma` is automatically driven by:

        1. **Worker dependencies**: Workers are scheduled to run only after all their necessary 
        dependencies are satisfied. The dependencies automatically drives the execution order.

        2. **Calling ferry_to**: During execution, a worker can explicitly trigger another worker 
        with calling `ferry_to()`, which enables dynamic flow control and conditional branching.

        3. **Dynamic topology changes**: When the graph topology is modified at runtime (such as adding 
        or removing workers or dependencies), the scheduling system seamlessly updates to reflect 
        the latest structure, ensuring that worker execution always follows the current graph.
        """

        def _reinit_current_kickoff_workers_if_needed():
            # Note: After deserialization, the _current_kickoff_workers must not be empty!
            # Therefore, _current_kickoff_workers will only be reinitialized when the Automa is run for the first time or rerun.
            # It is guaranteed that _current_kickoff_workers will not be reinitialized when the Automa is resumed after deserialization.
            if not self._current_kickoff_workers:
                self._current_kickoff_workers = [
                    _KickoffInfo(
                        worker_key=worker_key,
                        last_kickoff="__automa__"
                    ) for worker_key, worker_obj in self._workers.items()
                    if getattr(worker_obj, "is_start", False)
                ]
                # Each time the Automa re-runs, buffer the input arguments here.
                self._input_buffer.args = args
                self._input_buffer.kwargs = kwargs

        def _execute_topology_change_deferred_tasks(tc_tasks: List[Union[_AddWorkerDeferredTask, _RemoveWorkerDeferredTask, _AddDependencyDeferredTask]]):
            # update the control flow topology
            for topology_task in tc_tasks:
                if topology_task.task_type == "add_worker":
                    self._add_worker_incrementally(
                        key=topology_task.worker_key,
                        worker=topology_task.worker_obj,
                        dependencies=topology_task.dependencies,
                        is_start=topology_task.is_start,
                        is_output=topology_task.is_output,
                        args_mapping_rule=topology_task.args_mapping_rule,
                        result_dispatching_rule=topology_task.result_dispatching_rule,
                        callback_builders=topology_task.callback_builders,
                    )
                elif topology_task.task_type == "remove_worker":
                    self._remove_worker_incrementally(topology_task.worker_key)
                elif topology_task.task_type == "add_dependency":
                    self._add_dependency_incrementally(topology_task.worker_key, topology_task.dependency)

            # update the data flow topology
            args_manager.update_data_flow_topology(dynamic_tasks=tc_tasks)

        def _set_worker_run_finished(worker_key: str):
            for kickoff_info in self._current_kickoff_workers:
                if kickoff_info.worker_key == worker_key:
                    kickoff_info.run_finished = True
                    break

        def _check_and_normalize_interaction_params(
            feedback_data: Optional[Union[InteractionFeedback, List[InteractionFeedback]]] = None,
            interaction_feedback: Optional[InteractionFeedback] = None,
            interaction_feedbacks: Optional[List[InteractionFeedback]] = None,
        ):
            if feedback_data:
                if isinstance(feedback_data, list):
                    rx_feedbacks = feedback_data
                else:
                    rx_feedbacks = [feedback_data]
                return rx_feedbacks
            # For backward compatibility with old parameter names. To be removed in the future.
            if interaction_feedback and interaction_feedbacks:
                raise AutomaRuntimeError(
                    f"Only one of interaction_feedback or interaction_feedbacks can be used. "
                    f"But received interaction_feedback={interaction_feedback} and \n"
                    f"interaction_feedbacks={interaction_feedbacks}"
                )
            if interaction_feedback:
                rx_feedbacks = [interaction_feedback]
            else:
                rx_feedbacks = interaction_feedbacks
            return rx_feedbacks

        def _match_ongoing_interaction_and_feedbacks(rx_feedbacks:List[InteractionFeedback]):
            match_left_feedbacks = []
            for feedback in rx_feedbacks:
                matched = False
                for interaction_and_feedbacks in self._ongoing_interactions.values():
                    for interaction_and_feedback in interaction_and_feedbacks:
                        if interaction_and_feedback.interaction.interaction_id == feedback.interaction_id:
                            matched = True
                            # Note: Only one feedback is allowed for each interaction. Here we assume that only the first feedback is valid, which is a choice of implementation.
                            if interaction_and_feedback.feedback is None:
                                # Set feedback to self._ongoing_interactions
                                interaction_and_feedback.feedback = feedback
                            break
                    if matched:
                        break
                if not matched:
                    match_left_feedbacks.append(feedback)
            return match_left_feedbacks

        running_options = self._get_top_running_options()

        self._main_loop = asyncio.get_running_loop()
        self._main_thread_id = threading.get_ident()

        if self.thread_pool is None:
            self.thread_pool = ThreadPoolExecutor(thread_name_prefix="bridgic-thread")

        if not self._automa_running:
            # Here is the last chance to compile and check the DDG in the end of the [Initialization Phase] (phase 1 just before the first DS).
            self._compile_graph_and_detect_risks()
            self._automa_running = True

        # An Automa needs to be re-run with _current_kickoff_workers reinitialized.
        _reinit_current_kickoff_workers_if_needed()

        is_top_level = self.is_top_level()

        # If this is the top-level automa, execute its callbacks separately.
        if is_top_level:
            automa_callbacks = self._get_automa_callbacks()

            for callback in automa_callbacks:
                await callback.on_worker_start(
                    key=self.name,
                    is_top_level=True,
                    parent=self.parent,
                    arguments={
                        "args": self._input_buffer.args,
                        "kwargs": self._input_buffer.kwargs,
                        "feedback_data": feedback_data,
                    },
                )

        # For backward compatibility with old parameter names. To be removed in the future.
        interaction_feedback = kwargs.get("interaction_feedback")
        interaction_feedbacks = kwargs.get("interaction_feedbacks")
        rx_feedbacks = _check_and_normalize_interaction_params(feedback_data, interaction_feedback, interaction_feedbacks)
        if rx_feedbacks:
            rx_feedbacks = _match_ongoing_interaction_and_feedbacks(rx_feedbacks)

        if running_options.debug:
            printer.print(f"\n{type(self).__name__}-[{self.name}] is getting started.", color="green")

        # Task loop divided into many dynamic steps (DS).
        args_manager = ArgsManager(
            input_args=self._input_buffer.args,
            input_kwargs=self._input_buffer.kwargs,
            worker_outputs=self._worker_output,
            worker_forwards=self._worker_forwards,
            worker_dict=self._workers
        )
        is_output_worker_keys = set()
        while self._current_kickoff_workers:
            # A new DS started.
            if running_options.debug:
                kickoff_worker_keys = [kickoff_info.worker_key for kickoff_info in self._current_kickoff_workers]
                printer.print(f"[DS][Before Tasks Started] kickoff workers: {kickoff_worker_keys}", color="purple")

            for kickoff_info in self._current_kickoff_workers:
                if kickoff_info.run_finished:
                    # Skip finished workers. Here is the case that the Automa is resumed after a human interaction.
                    if running_options.debug:
                        printer.print(f"[{kickoff_info.worker_key}] will be skipped - run finished", color="blue")
                    continue

                if running_options.debug:
                    kickoff_name = kickoff_info.last_kickoff
                    if kickoff_name == "__automa__":
                        kickoff_name = f"{kickoff_name}:({self.name})"
                    printer.print(f"[{kickoff_name}] will kick off [{kickoff_info.worker_key}]", color="cyan")

                # Arguments Mapping:
                binding_args, binding_kwargs = args_manager.args_binding(
                    last_worker_key=kickoff_info.last_kickoff,
                    current_worker_key=kickoff_info.worker_key
                ) if not kickoff_info.from_ferry else ((), {})
                # Inputs Propagation
                _, propagation_kwargs = args_manager.inputs_propagation(current_worker_key=kickoff_info.worker_key)
                # Data injection.
                _, injection_kwargs = args_manager.args_injection(
                    current_worker_key=kickoff_info.worker_key, 
                    current_automa=self
                )
                # Ferry arguments.
                ferry_args, ferry_kwargs = kickoff_info.args, kickoff_info.kwargs
                # combine the arguments from the three steps.
                # kwargs will cover priority follows: propagation_kwargs < binding_kwargs < injection_kwargs < ferry_kwargs
                next_args, next_kwargs = safely_map_args(
                    (*binding_args, *ferry_args), 
                    {**propagation_kwargs, **binding_kwargs, **injection_kwargs, **ferry_kwargs}, 
                    self._workers[kickoff_info.worker_key].get_input_param_names()
                )
                
                # Collect the output worker keys.
                if self._workers[kickoff_info.worker_key].is_output:
                    is_output_worker_keys.add(kickoff_info.worker_key)
                    if len(is_output_worker_keys) > 1:
                        raise AutomaRuntimeError(
                            f"It is not allowed to have more than one worker with `is_output=True` and "
                            f"they are all considered as output-worker when the automa terminates and returns."
                            f"The current output-worker keys are: {is_output_worker_keys}."
                            f"If you want to collect the results of multiple workers simultaneously, "
                            f"it is recommended that you add one worker to gather them."
                        )

                # Schedule task for each kickoff worker.
                worker_obj = self._workers[kickoff_info.worker_key]
                if worker_obj.is_automa():
                    coro = worker_obj.arun(
                        *next_args,
                        feedback_data=rx_feedbacks,
                        **next_kwargs,
                    )
                else:
                    coro = worker_obj.arun(*next_args, **next_kwargs)

                task = asyncio.create_task(
                    # TODO1: arun() may need to be wrapped to support better interrupt...
                    coro,
                    name=f"Task-{kickoff_info.worker_key}"
                )
                self._running_tasks.append(_RunnningTask(
                    worker_key=kickoff_info.worker_key,
                    task=task,
                ))

            # Wait until all of the tasks are finished.
            while True:
                undone_tasks = [t.task for t in self._running_tasks if not t.task.done()]
                if not undone_tasks:
                    break
                try:
                    await undone_tasks[0]
                except Exception as e:
                    ...
                    # The same exception will be raised again in the following task.result().
                    # Note: A Task is done when the wrapped coroutine either returned a value, raised an exception, or the Task was cancelled.
                    # Refer to: https://docs.python.org/3/library/asyncio-task.html#task-object

            # Process graph topology change deferred tasks triggered by add_worker() and remove_worker().
            _execute_topology_change_deferred_tasks(self._topology_change_deferred_tasks)

            # Handle exceptions raised by all running tasks.
            interaction_exceptions: List[_InteractionEventException] = []
            non_interaction_exceptions: List[Exception] = []

            for task in self._running_tasks:
                try:
                    # It will raise an exception if task failed.
                    task_result = task.task.result()
                    _set_worker_run_finished(task.worker_key)

                    if task.worker_key in self._workers:
                        # The current running worker may be removed.
                        worker_obj = self._workers[task.worker_key]
                        # Collect results of the finished tasks.
                        self._worker_output[task.worker_key] = task_result
                        # reset dynamic states of finished workers.
                        self._workers_dynamic_states[task.worker_key].dependency_triggers = set(getattr(worker_obj, "dependencies", []))
                        # Update the dynamic states of successor workers.
                        for successor_key in self._worker_forwards.get(task.worker_key, []):
                            self._workers_dynamic_states[successor_key].dependency_triggers.remove(task.worker_key)
                        # Each time a worker is finished running, the ongoing interaction states should be cleared. Once it is re-run, the human interactions in the worker can be triggered again.
                        if task.worker_key in self._worker_interaction_indices:
                            del self._worker_interaction_indices[task.worker_key]
                        if task.worker_key in self._ongoing_interactions:
                            del self._ongoing_interactions[task.worker_key]
                except Exception as e:
                    if isinstance(e, _InteractionEventException):
                        interaction_exceptions.append(e)
                        if task.worker_key in self._workers and not self._workers[task.worker_key].is_automa():
                            if task.worker_key not in self._ongoing_interactions:
                                self._ongoing_interactions[task.worker_key] = []
                            interaction=e.args[0]
                            # Make sure the interaction_id is unique for each human interaction.
                            found = False
                            for iaf in self._ongoing_interactions[task.worker_key]:
                                if iaf.interaction.interaction_id == interaction.interaction_id:
                                    found = True
                                    break
                            if not found:
                                self._ongoing_interactions[task.worker_key].append(_InteractionAndFeedback(
                                    interaction=interaction,
                                ))
                    else:
                        non_interaction_exceptions.append(e)

            if len(self._topology_change_deferred_tasks) > 0:
                # Graph topology validation and risk detection. Only needed when topology changes.
                # Guarantee the graph topology is valid and consistent after each DS.
                # 1. Validate the canonical graph.
                self._validate_canonical_graph()
                # 2. Validate the DAG constraints.
                GraphMeta.validate_dag_constraints(self._worker_forwards)
                # TODO: more validations can be added here...

            # TODO: Ferry-related risk detection may be added here...

            # Handle exceptions with callbacks at the top-level automa before re-raising them.
            if is_top_level:
                # Get cached callbacks for top-level automa
                automa_callbacks = self._get_automa_callbacks()

                # Process interaction exceptions with callbacks (they cannot be suppressed, but callbacks can observe them)
                for e in interaction_exceptions + non_interaction_exceptions:
                    await try_handle_error_with_callbacks(
                        callbacks=automa_callbacks,
                        key=self.name,
                        is_top_level=True,
                        parent=self.parent,
                        arguments={
                            "args": self._input_buffer.args,
                            "kwargs": self._input_buffer.kwargs,
                            "feedback_data": feedback_data,
                        },
                        error=e,
                    )

            # For inner interaction exceptions, collect them and throw an InteractionException as a whole.
            if len(interaction_exceptions) > 0:
                all_interactions: List[Interaction] = [interaction for e in interaction_exceptions for interaction in e.args]
                if self.is_top_level():
                    # This is the top-level Automa. Serialize the Automa and raise InteractionException to the application layer.
                    serialized_automa = dump_bytes(self)
                    snapshot = Snapshot(
                        serialized_bytes=serialized_automa,
                        serialization_version=GraphAutoma.SERIALIZATION_VERSION,
                    )
                    raise InteractionException(
                        interactions=all_interactions,
                        snapshot=snapshot,
                    )
                else:
                    # Continue raise exception to the upper level Automa.
                    raise _InteractionEventException(*all_interactions)

            # For non-interaction exceptions, immediately raise the first one directly, since none of them are meant to be suppressed.
            if len(non_interaction_exceptions) > 0:
                raise non_interaction_exceptions[0]

            # Find next kickoff workers and rebuild _current_kickoff_workers
            run_finished_worker_keys: List[str] = [kickoff_info.worker_key for kickoff_info in self._current_kickoff_workers if kickoff_info.run_finished]
            assert len(run_finished_worker_keys) == len(self._current_kickoff_workers)
            self._current_kickoff_workers = []
            # New kickoff workers can be triggered by two ways:
            # 1. The ferry_to() operation is called during current worker execution.
            # 2. The dependencies are eliminated after all predecessor workers are finished.
            # So,
            # First add kickoff workers triggered by ferry_to();
            for ferry_task in self._ferry_deferred_tasks:
                self._current_kickoff_workers.append(_KickoffInfo(
                    worker_key=ferry_task.ferry_to_worker_key,
                    last_kickoff=ferry_task.kickoff_worker_key,
                    from_ferry=True,
                    args=ferry_task.args,
                    kwargs=ferry_task.kwargs,
                ))
            # Then add kickoff workers triggered by dependencies elimination.
            # Merge successor keys of all finished tasks.
            successor_keys = set()
            for worker_key in run_finished_worker_keys:
                # Note: The `worker_key` worker may have been removed from the Automa.
                for successor_key in self._worker_forwards.get(worker_key, []):
                    if successor_key not in successor_keys:
                        dependency_triggers = self._workers_dynamic_states[successor_key].dependency_triggers
                        if not dependency_triggers:
                            self._current_kickoff_workers.append(_KickoffInfo(
                                worker_key=successor_key,
                                last_kickoff=worker_key,
                            ))
                        successor_keys.add(successor_key)
            if running_options.debug:
                deferred_ferrys = [ferry_task.ferry_to_worker_key for ferry_task in self._ferry_deferred_tasks]
                printer.print(f"[DS][After Tasks Finished] successor workers: {successor_keys}, deferred ferrys: {deferred_ferrys}", color="purple")

            # Clear running tasks after all finished.
            self._running_tasks.clear()
            self._ferry_deferred_tasks.clear()
            self._topology_change_deferred_tasks.clear()

        if running_options.debug:
            printer.print(f"{type(self).__name__}-[{self.name}] is finished.", color="green")

        # After a complete run, reset all necessary states to allow the automa to re-run.
        self._input_buffer = _AutomaInputBuffer()
        if self.should_reset_local_space():
            self._clean_all_worker_local_space()
        self._ongoing_interactions.clear()
        self._worker_interaction_indices.clear()
        self._automa_running = False

        # Get result before calling callbacks
        if is_output_worker_keys:
            result = self._worker_output.get(list(is_output_worker_keys)[0], None)
        else:
            result = None

        # If this is the top-level automa, execute its callbacks separately.
        if is_top_level:
            automa_callbacks = self._get_automa_callbacks()
            for callback in automa_callbacks:
                await callback.on_worker_end(
                    key=self.name,
                    is_top_level=True,
                    parent=self.parent,
                    arguments={
                        "args": self._input_buffer.args,
                        "kwargs": self._input_buffer.kwargs,
                        "feedback_data": feedback_data,
                    },
                    result=result,
                )

        return result

    def _get_worker_dependencies(self, worker_key: str) -> List[str]:
        """
        Get the worker keys of all dependencies of the worker.
        """
        deps = self._workers[worker_key].dependencies
        return [] if deps is None else deps

    def _find_connected_components(self):
        """
        Find all of the connected components in the whole automa graph described by self._workers.
        """
        visited = set()
        component_list = []
        component_idx = {}

        def dfs(worker: str, component: List[str]):
            visited.add(worker)
            component.append(worker)
            for target in self._worker_forwards.get(worker, []):
                if target not in visited:
                    dfs(target, component)

        for worker in self._workers.keys():
            if worker not in visited:
                component_list.append([])
                current_idx = len(component_list) - 1
                current_component = component_list[current_idx]

                dfs(worker, current_component)

                for worker in current_component:
                    component_idx[worker] = current_idx

        # self._component_list, self._component_idx = component_list, component_idx
        # TODO: check how to use _component_list and _component_idx...

    @override
    def _get_worker_key(self, worker: Worker) -> Optional[str]:
        for worker_key, worker_obj in self._workers.items():
            if worker_obj == worker:
                # Note: _GraphAdaptedWorker.__eq__() is overridden to support the '==' operator.
                return worker_key
        return None

    @override
    def _get_worker_instance(self, worker_key: str) -> Worker:
        return self._workers[worker_key]

    @override
    def _locate_interacting_worker(self) -> Optional[str]:
        return self._trace_back_kickoff_worker_key_from_stack()

    def _trace_back_kickoff_worker_key_from_stack(self) -> Optional[str]:
        worker = self._get_current_running_worker_instance_by_stacktrace()
        if worker:
            return self._get_worker_key(worker)
        return None

    def _get_current_running_worker_instance_by_stacktrace(self) -> Optional[Worker]:
        for frame_info in inspect.stack():
            frame = frame_info.frame
            if 'self' in frame.f_locals:
                self_obj = frame.f_locals['self']
                if isinstance(self_obj, Worker) and (not isinstance(self_obj, Automa)) and (frame_info.function == "arun" or frame_info.function == "run"):
                    return self_obj
        return None

    def __repr__(self) -> str:
        # TODO : It's good to make __repr__() of Automa compatible with eval().
        # This feature depends on the implementation of __repr__() of workers.
        class_name = self.__class__.__name__
        workers_str = self._workers.__repr__()
        return f"{class_name}(workers={workers_str})"

    def __str__(self) -> str:
        d = {}
        for k, v in self._workers.items():
            d[k] = f"{v} depends on {getattr(v, 'dependencies', [])}"
        return json.dumps(d, ensure_ascii=False, indent=4)



import inspect
from typing import Callable, Dict, Any, List, Optional, cast, get_origin, Annotated
from pydantic import create_model, WithJsonSchema, Field, ConfigDict
from pydantic.fields import FieldInfo
from bridgic.core.automa.args._args_descriptor import JSON_SCHEMA_IGNORE_ARG_TYPES
from docstring_parser import parse as parse_docstring # type: ignore

def create_func_params_json_schema(
    func: Callable,
    ignore_params: List[str] = ["self", "cls"],
) -> Dict[str, Any]:
    """
    Create a JSON schema for the parameters of a function. The docstring of the function will be used to extract the description of the parameters, if available.

    Parameters
    ----------
    func : Callable
        The function to create a JSON schema for.
    ignore_params : List[str]
        The parameters to ignore, not included in the resulting JSON schema.

    Returns
    -------
    Dict[str, Any]
        The JSON schema for the parameters of the function.
    """
    def _get_param_description_from_annotation(annotated_type: Any) -> Optional[str]:
        for metadata in annotated_type.__metadata__:
            if isinstance(metadata, FieldInfo):
                return metadata.description
            if isinstance(metadata, WithJsonSchema) and metadata.json_schema:
                return metadata.json_schema.get("description")
        return None

    sig = inspect.signature(func)
    docstring = parse_docstring(func.__doc__)
    field_defs: Dict[str, Any] = {}

    for param_name, param in sig.parameters.items():
        if param_name in ignore_params:
            continue

        # First: Resolve the type of the parameter
        param_type = param.annotation
        if param_type is inspect.Parameter.empty:
            # In the case of no annotation
            param_type = Any

        # Second: Resolve the default of the parameter
        param_default = param.default
        if isinstance(param_default, JSON_SCHEMA_IGNORE_ARG_TYPES):
            continue
        # param_default may be inspect.Parameter.empty in the case of no default.
    
        # Third: Resolve the description of the parameter
        # The description of a parameter may come from multiple sources, with the following order of priority (from highest to lowest):
        # 1. param: type = Field(description="description here", ...)
        # 2. param: Annotated[type, "description here"]
        # 3. param: Annotated[type, Field(description="description here", ...)]
        # 4. param: Annotated[type, WithJsonSchema({"description":"description here", ...})]
        # 5. the params section of the function's docstring
        param_description = None
        need_overwritten = False # Whether the description of the parameter needs to be overwritten.
        if isinstance(param_default, FieldInfo):
            param_description = param_default.description
        if not param_description and get_origin(param_type) is Annotated:
            if len(param_type.__metadata__) == 1 and isinstance(param_type.__metadata__[0], str):
                param_description = cast(str, param_type.__metadata__[0])
                need_overwritten = bool(param_description)
            else:
                param_description = _get_param_description_from_annotation(param_type)
        if not param_description and docstring.params:
            # Search the param description in the docstring.params
            for p in docstring.params:
                if p.arg_name == param_name:
                    param_description = p.description
                    need_overwritten = bool(param_description)
                    break
        
        # Assemble the Field object and add it to the field_defs
        if need_overwritten:
            if param_default is inspect.Parameter.empty:
                field = Field(description=param_description)
            elif isinstance(param_default, FieldInfo):
                field = param_default
                field.description = param_description
            else:
                field = Field(default=param_default, description=param_description)
            field_defs[param_name] = (param_type, field)
        else:
            if param_default is inspect.Parameter.empty:
                field_defs[param_name] = param_type
            else:
                field_defs[param_name] = (param_type, param_default)
    
    # Note: Set arbitrary_types_allowed=True to allow custom parameter types that support JSON schema, by implementting `__get_pydantic_core_schema__` or `__get_pydantic_json_schema__`.
    JsonSchemaModel = create_model(
        func.__name__,
        __config__=ConfigDict(arbitrary_types_allowed=True), 
        **field_defs
    )
    return JsonSchemaModel.model_json_schema()


from typing import Optional, Union, Dict, Any
from abc import abstractmethod
from bridgic.core.types._serialization import Serializable
from bridgic.core.model.types import Tool
from bridgic.core.automa.worker import Worker
from typing_extensions import override

class ToolSpec(Serializable):
    """
    ToolSpec is an abstract class that represents a tool specification that describes all necessary information about a tool used by the LLM. 

    ToolSpec and its subclasses are responsible for providing four categories of interfaces:
    1. Transformations to LLM Tool: `to_tool`.
    2. Worker Creation: `create_worker`.
    3. Serialization and Deserialization.
    4. ToolSpec initialization from raw resources: `from_raw`.
    """
    _tool_id: Optional[Union[str, int]]
    """The unique ID of the tool, used to uniquely identify a tool across the entire system. This tool can be of various types."""
    _tool_name: Optional[str]
    """The name of the tool to be called"""
    _tool_description: Optional[str]
    """A description of what the tool does, used by the model to choose when and how to call the tool."""
    _tool_parameters: Optional[Dict[str, Any]]
    """The JSON schema of the tool's parameters"""

    def __init__(
        self,
        tool_name: Optional[str] = None,
        tool_description: Optional[str] = None,
        tool_parameters: Optional[Dict[str, Any]] = None,
    ):
        self._tool_id = None
        self._tool_name = tool_name
        self._tool_description = tool_description
        self._tool_parameters = tool_parameters

    @property
    def tool_name(self) -> Optional[str]:
        return self._tool_name

    @property
    def tool_description(self) -> Optional[str]:
        return self._tool_description

    @property
    def tool_parameters(self) -> Optional[Dict[str, Any]]:
        return self._tool_parameters

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(tool_name={self._tool_name}, tool_description={self._tool_description}, tool_parameters={self._tool_parameters})"
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(tool_name={self._tool_name}, tool_description={self._tool_description}, tool_parameters={self._tool_parameters})>"

    ###############################################################
    ######## Part One of interfaces: Transformations to Tool ######
    ###############################################################

    @abstractmethod
    def to_tool(self) -> Tool:
        """
        Transform this ToolSpec to a `Tool` object used by LLM.

        Returns
        -------
        Tool
            A `Tool` object that can be used by LLM.
        """
        ...

    ###############################################################
    ######## Part Two of interfaces: Worker Creation ##############
    ###############################################################

    @abstractmethod
    def create_worker(self) -> Worker:
        """
        Create a Worker from the information included in this ToolSpec.

        Returns
        -------
        Worker
            A new `Worker` object that can be added to an Automa to execute the tool.
        """
        ...

    ###############################################################
    ######## Part Three of interfaces: 
    ######## Serialization and Deserialization ####################
    ###############################################################

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = {}
        if self._tool_id:
            state_dict["tool_id"] = self._tool_id
        if self._tool_name:
            state_dict["tool_name"] = self._tool_name
        if self._tool_description:
            state_dict["tool_description"] = self._tool_description
        if self._tool_parameters:
            state_dict["tool_parameters"] = self._tool_parameters
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        self._tool_id = state_dict.get("tool_id")
        self._tool_name = state_dict.get("tool_name")
        self._tool_description = state_dict.get("tool_description")
        self._tool_parameters = state_dict.get("tool_parameters")

    ###############################################################
    ######## Part Four of interfaces: 
    ######## ToolSpec initialization from raw resources ###########
    ######## `from(...)`: See subclasses for details ##############
    ###############################################################



import inspect
from typing import Optional, Union, Dict, List, Any, Type, Callable, Annotated, get_origin
from typing_extensions import override
from abc import abstractmethod
from types import MethodType
from docstring_parser import parse as parse_docstring # type: ignore

from bridgic.core.model.types import Tool
from bridgic.core.automa import Automa
from bridgic.core.automa.worker import Worker
from bridgic.core.agentic.tool_specs._base_tool_spec import ToolSpec
from bridgic.core.utils._json_schema import create_func_params_json_schema
from bridgic.core.utils._inspect_tools import load_qualified_class_or_func

def as_tool(spec_func: Callable) -> Callable:
    """
    A decorator that transforms a class to a tool that may be used by LLM.

    Parameters
    ----------
    spec_func : Callable
        The function used to declare the tool spec. Note that this function is not intended to be called directly.
    """
    def decorator(cls):
        if not isinstance(spec_func, Callable):
            raise ValueError(f"A function argument is expected, but got {type(spec_func)}.")
        if isinstance(spec_func, MethodType):
            raise ValueError(f"`spec_func` is not allowed to be a bound method: {spec_func}.")
        cls.spec_func = spec_func
        return cls
    return decorator

def get_tool_description_from(
    spec_func: Callable,
    tool_name: Optional[str] = None
) -> str:
    """
    Get the tool description from the spec function.
    """
    docstring = parse_docstring(spec_func.__doc__)
    tool_description = docstring.description
    if tool_description:
        tool_description = tool_description.strip()

    if not tool_description:
        # No description provided, use the function signature as the description.
        fn_sig = inspect.signature(spec_func)
        filtered_params = []
        ignore_params: List[str] = ["self", "cls"]
        for param_name, param in fn_sig.parameters.items():
            if param_name in ignore_params:
                continue

            # Resolve the original type of the parameter.
            param_type = param.annotation
            if get_origin(param_type) is Annotated:
                param_type = param_type.__origin__

            # Note: Remove the default of the parameter.
            filtered_params.append(param.replace(
                annotation=param_type,
                default=inspect.Parameter.empty
            ))

        fn_sig = fn_sig.replace(parameters=filtered_params)
        if not tool_name:
            tool_name = spec_func.__name__
        tool_description = f"{tool_name}{fn_sig}\n"
    return tool_description

class AutomaToolSpec(ToolSpec):
    _automa_cls: Type[Automa]
    """The Automa class to be used as a tool"""
    _automa_init_kwargs: Dict[str, Any]
    """The initialization arguments for the Automa"""

    def __init__(
        self,
        automa_cls: Type[Automa],
        tool_name: Optional[str] = None,
        tool_description: Optional[str] = None,
        tool_parameters: Optional[Dict[str, Any]] = None,
        **automa_init_kwargs: Dict[str, Any],
    ):
        super().__init__(
            tool_name=tool_name,
            tool_description=tool_description,
            tool_parameters=tool_parameters
        )
        self._automa_cls = automa_cls
        self._automa_init_kwargs = automa_init_kwargs

    @classmethod
    def from_raw(
        cls,
        automa_cls: Type[Automa],
        tool_name: Optional[str] = None,
        tool_description: Optional[str] = None,
        tool_parameters: Optional[Dict[str, Any]] = None,
        **automa_init_kwargs: Dict[str, Any],
    ) -> "AutomaToolSpec":
        """
        Create an AutomaToolSpec from an Automa class.
        """

        def check_spec_func(automa_cls):
            if hasattr(automa_cls, "spec_func") and isinstance(automa_cls.spec_func, Callable):
                return
            raise ValueError(f"The Automa class {automa_cls} must be decorated with `@as_tool` in order to be used as a tool.")

        if (not tool_name) or (not tool_description) or (not tool_parameters):
            check_spec_func(automa_cls)

        if not tool_name:
            tool_name = automa_cls.spec_func.__name__
        
        if not tool_description:
            tool_description = get_tool_description_from(automa_cls.spec_func, tool_name)

        if not tool_parameters:
            tool_parameters = create_func_params_json_schema(automa_cls.spec_func)
            # TODO: whether to remove the `title` field of the params_schema?

        return cls(
            automa_cls=automa_cls,
            tool_name=tool_name,
            tool_description=tool_description,
            tool_parameters=tool_parameters,
            **automa_init_kwargs
        )

    @override
    def to_tool(self) -> Tool:
        """
        Transform this AutomaToolSpec to a `Tool` object used by LLM.

        Returns
        -------
        Tool
            A `Tool` object that can be used by LLM.
        """
        return Tool(
            name=self._tool_name,
            description=self._tool_description,
            parameters=self._tool_parameters
        )

    @override
    def create_worker(self) -> Worker:
        """
        Create a Worker from the information included in this AutomaToolSpec.

        Returns
        -------
        Worker
            A new `Worker` object that can be added to an Automa to execute the tool.
        """
        return self._automa_cls(**self._automa_init_kwargs)

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()
        state_dict["automa_cls"] = self._automa_cls.__module__ + "." + self._automa_cls.__qualname__
        if self._automa_init_kwargs:
            state_dict["automa_init_kwargs"] = self._automa_init_kwargs
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)
        self._automa_cls = load_qualified_class_or_func(state_dict["automa_cls"])
        self._automa_init_kwargs = state_dict.get("automa_init_kwargs") or {}




from threading import Lock
from typing import Dict, Union, Literal, Callable

legal_colors = ["red", "green", "yellow", "blue", "purple", "cyan", "white", "magenta"]

color_funcs: Dict[str, Callable[[str], str]] = {
    "red": lambda msg: "\033[91m{}\033[00m".format(msg),
    "green": lambda msg: "\033[92m{}\033[00m".format(msg), 
    "yellow": lambda msg: "\033[93m{}\033[00m".format(msg),
    "blue": lambda msg: "\033[94m{}\033[00m".format(msg),
    "purple": lambda msg: "\033[95m{}\033[00m".format(msg),
    "cyan": lambda msg: "\033[96m{}\033[00m".format(msg),
    "white": lambda msg: "\033[97m{}\033[00m".format(msg),
    "magenta": lambda msg: "\033[35m{}\033[00m".format(msg)
}

def colored(
    msg: str,
    color: Union[Literal["red", "green", "yellow", "blue", "purple", "cyan", "white", "magenta"], None] = None,
) -> str:
    if color not in legal_colors:
        raise ValueError(f"Invalid color: {color}")
    return color_funcs[color](msg)

class Printer:
    """
    A thread-safe printer class that can print colored text to the console.

    Parameters
    ----------
    color : Union[Literal["red", "green", "yellow", "blue", "purple", "cyan", "white", "magenta"], None]
        The color of the text. If None, the text will be printed in the default color.
    """
    lock: Lock = Lock()

    @classmethod
    def print(
        cls,
        *values,
        color: Union[Literal["red", "green", "yellow", "blue", "purple", "cyan", "white", "magenta"], None] = None,
        **kwargs,
    ):
        if color:
            if color not in legal_colors:
                raise ValueError(f"Invalid color: {color}")
            values = list(map(lambda x: color_funcs[color](x), values))

        with cls.lock:
            print(*values, **kwargs, flush=True)

printer = Printer()


import warnings
import sys

from threading import Lock
from typing import Any, Dict, Union, List, Optional, Type, Generic, TypeVar, TYPE_CHECKING, get_origin, get_args, get_type_hints
from typing_extensions import override
from bridgic.core.types._serialization import Serializable
from bridgic.core.utils._inspect_tools import load_qualified_class_or_func

if TYPE_CHECKING:
    from bridgic.core.automa._automa import Automa

T_WorkerCallback = TypeVar("T_WorkerCallback", bound="WorkerCallback")
"""Type variable for WorkerCallback subclasses."""


class WorkerCallback(Serializable):
    """
    Callback for the execution of a worker instance during the running 
    of a prebult automa.

    This class defines the interfaces that will be called before or after 
    the execution of the corresponding worker. Callbacks are typically used 
    for validating input, monitoring execution, and collecting logs, etc.

    Methods
    -------
    on_worker_start(key, is_top_level, parent, arguments)
        Hook invoked before worker execution.
    on_worker_end(key, is_top_level, parent, arguments, result)
        Hook invoked after worker execution.
    on_worker_error(key, is_top_level, parent, arguments, error)
        Hook invoked when worker execution raises an exception.
    """
    async def on_worker_start(
        self, 
        key: str,
        is_top_level: bool = False,
        parent: Optional["Automa"] = None,
        arguments: Dict[str, Any] = None,
    ) -> None:
        """
        Hook invoked before worker execution.

        Called immediately before the worker runs. Use for arguments
        validation, logging, or monitoring. Cannot modify execution
        arguments or logic.

        Parameters
        ----------
        key : str
            Worker identifier.
        is_top_level: bool = False
            Whether the worker is the top-level automa. When True, parent will be the automa itself (parent is self).
        parent : Optional[Automa] = None
            Parent automa instance containing this worker. For top-level automa, parent is the automa itself.
        arguments : Dict[str, Any] = None
            Execution parameters with keys "args" and "kwargs".
        """
        pass

    async def on_worker_end(
        self,
        key: str,
        is_top_level: bool = False,
        parent: Optional["Automa"] = None,
        arguments: Dict[str, Any] = None,
        result: Any = None,
    ) -> None:
        """
        Hook invoked after worker execution.

        Called immediately after the worker completes. Use for result
        monitoring, logging, event publishing, or validation. Cannot
        modify execution results or logic.

        Parameters
        ----------
        key : str
            Worker identifier.
        is_top_level: bool = False
            Whether the worker is the top-level automa. When True, parent will be the automa itself (parent is self).
        parent : Optional[Automa] = None
            Parent automa instance containing this worker. For top-level automa, parent is the automa itself.
        arguments : Dict[str, Any] = None
            Execution arguments with keys "args" and "kwargs".
        result : Any = None
            Worker execution result.
        """
        pass

    async def on_worker_error(
        self,
        key: str,
        is_top_level: bool = False,
        parent: Optional["Automa"] = None,
        arguments: Dict[str, Any] = None,
        error: Exception = None,
    ) -> bool:
        """
        Hook invoked when worker execution raises an exception.

        Called when the worker execution raises an exception. Use for error handling, logging, 
        or event publishing. Cannot modify execution logic or arguments.

        **Exception Matching Mechanism: How to Handle a Specific Exception**

        The framework enable your callback to handle a given exception based on the 
        **type annotation** of the `error` parameter in your `on_worker_error` method.
        The matching follows these rules:

        - The parameter name MUST be `error` and the type annotation is critical for the 
          matching mechanism.
        - If you annotate `error: ValueError`, it will match `ValueError` and all its 
          subclasses (e.g., `UnicodeDecodeError`).
        - If you annotate `error: Exception`, it will match all exceptions (since all exceptions 
          inherit from Exception).
        - If you want to match multiple exception types, you can use `Union[Type1, Type2, ...]`.


        **Return Value: Whether to Suppress the Exception**

        - If `on_worker_error` returns `True`, the framework will suppress the exception. 
          The framework will then proceed as if there was no error, and the worker result 
          will be set to None.
        - If `on_worker_error` returns `False`, the framework will simply observe the error; 
          after all matching callbacks are called, the framework will re-raise the exception.

        **Special Case: Interaction Exceptions Cannot Be Suppressed**

        To ensure human-interaction mechanisms work correctly, exceptions of type
        `_InteractionEventException` or `InteractionException` (including their subclasses) 
        **CANNOT** be suppressed by any callback. Even if your callback returns `True`, the 
        framework will forcibly re-raise the exception. This ensures these exceptions always 
        propagate correctly through the automa hierarchy to trigger necessary human interactions.

        Parameters
        ----------
        key : str
            Worker identifier.
        is_top_level: bool = False
            Whether the worker is the top-level automa. When True, parent will be the automa itself (parent is self).
        parent : Optional[Automa] = None
            Parent automa instance containing this worker. For top-level automa, parent is the automa itself.
        arguments : Dict[str, Any] = None
            Execution arguments with keys "args" and "kwargs".
        error : Exception = None
            The exception raised during worker execution. The type annotation of this
            parameter determines which exceptions this callback will handle. The matching
            is based on inheritance relationship (using isinstance), so a callback with
            `error: ValueError` will match ValueError and all its subclasses.

        Returns
        -------
        bool
            True if the automa should suppress the exception (not re-raise it); False otherwise.
        """
        return False

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        return {
            "callback_cls": self.__class__.__module__ + "." + self.__class__.__qualname__,
        }

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        pass

class WorkerCallbackBuilder(Generic[T_WorkerCallback]):
    """
    Builder class for creating instances of `WorkerCallback` subclasses.

    This builder is designed to construct instances of subclasses of `WorkerCallback`.
    The `_callback_type` parameter should be a subclass of `WorkerCallback`, and `build()` 
    will return an instance of that specific subclass. There is no need to call `build()` 
    directly. Instead, the framework calls the `build` method automatically to create 
    its own `WorkerCallback` instance for each worker instance.

    Notes
    -----
    **Register a Callback in Different Scope**

    There are three ways to register a callback for three levels of customization:

    - Case 1: Use in worker decorator to register the callback for a specific worker.
    - Case 2: Use in RunningOptions to register the callback for a specific Automa instance.
    - Case 3: Use in GlobalSetting to register the callback for all workers.

    Notes
    -----
    **Shared Instance Mode**

    - When `is_shared=True` (default), all workers within the same scope will share the same 
      callback instance. This is useful for scenarios where a single callback instance is needed 
      to maintain some state across workers within the same scope, such as the connection to 
      an external service. The scope is determined by where the builder is declared:
      - If declared in GlobalSetting: shared across all workers globally
      - If declared in RunningOptions: shared across all workers within that Automa instance
    - When `is_shared=False`, each worker will get its own callback instance. This is useful for 
      scenarios where a independent callback instance is needed for each worker.

    Examples
    --------
    There are three ways to use the builder, for different levels of customization:

    >>> # Define a custom callback class:
    >>> class MyEmptyCallback(WorkerCallback):
    ...     pass
    ...
    >>> # Case 1: Use in worker decorator to register the callback for a specific worker:
    >>> class MyGraphAutoma(GraphAutoma):
    ...     @worker(callback_builders=[WorkerCallbackBuilder(MyEmptyCallback)])
    ...     async def my_worker(self, x: int) -> int:
    ...         return x + 1
    ...
    >>> # Case 2: Use in RunningOptions to register the callback for a specific Automa instance:
    ...     running_options = RunningOptions(callback_builders=[WorkerCallbackBuilder(MyEmptyCallback)])
    ...     graph = MyGraphAutoma(running_options=running_options)
    ...
    >>> # Case 3: Use in GlobalSetting to register the callback for all workers:
    >>> GlobalSetting.set(callback_builders=[WorkerCallbackBuilder(MyEmptyCallback)])
    """
    _callback_type: Type[T_WorkerCallback]
    """The specific subclass of `WorkerCallback` to instantiate."""
    _init_kwargs: Dict[str, Any]
    """The initialization arguments for the instance."""
    _is_shared: bool
    """Whether to use shared instance mode (reuse the same instance within the declaration scope)."""

    _shared_instance: Optional[T_WorkerCallback] = None
    """Shared instance of the callback within the declaration scope."""
    _shared_lock: Lock = Lock()
    """Lock for thread-safe shared instance creation."""

    def __init__(
        self,
        callback_type: Type[T_WorkerCallback],
        init_kwargs: Optional[Dict[str, Any]] = None,
        is_shared: bool = True,
    ):
        """
        Initialize the builder with a `WorkerCallback` subclass type.

        Parameters
        ----------
        callback_type : Type[T_WorkerCallback]
            A subclass of `WorkerCallback` to be instantiated.
        init_kwargs : Optional[Dict[str, Any]]
            Keyword arguments to pass to the subclass constructor.
        is_shared : bool, default True
            If True, the callback instance will be shared within the declaration scope:
            If False, each worker will get its own callback instance.
        """
        self._callback_type = callback_type
        self._init_kwargs = init_kwargs or {}
        self._is_shared = is_shared

    def build(self) -> T_WorkerCallback:
        """
        Build and return an instance of the specified `WorkerCallback` subclass.

        Returns
        -------
        T_WorkerCallback
            An instance of the `WorkerCallback` subclass specified during initialization.
        """
        if self._is_shared:
            if self._shared_instance is None:
                with self._shared_lock:
                    if self._shared_instance is None:
                        self._shared_instance = self._callback_type(**self._init_kwargs)
            return self._shared_instance
        else:
            return self._callback_type(**self._init_kwargs)

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        return {
            "callback_type": self._callback_type.__module__ + "." + self._callback_type.__qualname__,
            "init_kwargs": self._init_kwargs,
            "is_shared": self._is_shared,
        }

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        # Load the callback type from its fully qualified name
        callback_type_name = state_dict["callback_type"]
        self._callback_type = load_qualified_class_or_func(callback_type_name)

        # Load init_kwargs (default to empty dict if not present or None)
        init_kwargs = state_dict.get("init_kwargs")
        self._init_kwargs = init_kwargs if init_kwargs is not None else {}

        # Load is_shared (default to True if not present for backward compatibility)
        self._is_shared = state_dict.get("is_shared", True)

        # Reset shared instance and lock (they will be recreated when needed)
        self._shared_instance = None
        self._shared_lock = Lock()


def can_handle_exception(callback: WorkerCallback, error: Exception) -> bool:
    """
    Check if a callback can handle a specific exception type based on its type annotation.

    The matching is based on inheritance relationship using `isinstance()`, not exact
    type matching. This means:
    - A callback with `error: ValueError` will match `ValueError` and all its subclasses.
    - A callback with `error: Exception` will match all exceptions (since all exceptions inherit from Exception).
    - Union types are supported: `error: Union[ValueError, TypeError]` will match both.

    Parameters
    ----------
    callback : WorkerCallback
        The callback instance to check.
    error : Exception
        The exception to check.

    Returns
    -------
    bool
        True if the callback can handle this exception type (based on inheritance),
        False otherwise.
    """
    try:
        # Maintain the namespace dictionary.
        callback_module = sys.modules.get(callback.__class__.__module__)
        globalns = callback_module.__dict__ if callback_module else {}

        # Import Automa here and add it to the namespace to resolve forward references.
        from bridgic.core.automa._automa import Automa
        globalns = {**globalns, 'Automa': Automa}

        # Extract the error type.
        annotations = get_type_hints(callback.on_worker_error, globalns=globalns)
        error_type = annotations.get("error")
        if error_type is None:
            warnings.warn(
                f"No type annotation found for the `error` parametor of "
                f"`{callback.__class__.__name__}.on_worker_error`."
            )
            return False

        origin = get_origin(error_type)

        # Handle Union types (e.g., Union[ValueError, TypeError])
        if origin is Union:
            union_args = get_args(error_type)
            # Check if exception is instance of any type in the Union
            return any(isinstance(error, t) for t in union_args if isinstance(t, type) and issubclass(t, Exception))

        # Handle single type annotation (including base Exception class)
        if isinstance(error_type, type) and issubclass(error_type, Exception):
            # Match based on inheritance relationship using isinstance
            return isinstance(error, error_type)
    except (ValueError, TypeError, AttributeError, NameError):
        # If we can't determine the type, skip this callback
        pass
    
    return False


async def try_handle_error_with_callbacks(
    callbacks: List[WorkerCallback],
    key: str,
    is_top_level: bool = False,
    parent: Optional["Automa"] = None,
    arguments: Dict[str, Any] = None,
    error: Exception = None,
) -> bool:
    """
    Try to handle an exception using the provided callbacks.

    This function returns True if at least one callback's `on_worker_error` method returns True
    (indicating a request to suppress the exception) and the exception to be handled is not an 
    interaction exception. Otherwise the function returns False, which means the framework will 
    re-raise the exception as usual.

    **Important**: Interaction exceptions (`_InteractionEventException` or `InteractionException`
    and their subclasses) cannot be suppressed. If the exception is an interaction exception,
    this function will always return False regardless of callback return values.

    Parameters
    ----------
    callbacks : List[WorkerCallback]
        List of callbacks to check.
    key : str
        Worker identifier.
    is_top_level : bool, optional
        Whether the worker is the top-level automa. Default is False. When True, parent will be the automa itself (parent is self).
    parent : Optional[Automa], optional
        Parent automa instance containing this worker. For top-level automa, parent is the automa itself.
    arguments : Dict[str, Any], optional
        Execution arguments with keys "args" and "kwargs".
    error : Exception, optional
        The exception to handle.

    Returns
    -------
    bool
        True if at least one callback requested to suppress the exception and the exception
        is not an interaction exception; False otherwise. The framework will re-raise the 
        exception if this returns False.
    """
    # Import here to avoid circular import
    from bridgic.core.automa._automa import _InteractionEventException
    from bridgic.core.automa.interaction._human_interaction import InteractionException
    
    should_suppress = False
    is_interaction_exception = isinstance(error, (_InteractionEventException, InteractionException))

    for callback in callbacks:
        if can_handle_exception(callback, error):
            suppress_request = await callback.on_worker_error(
                key=key,
                is_top_level=is_top_level,
                parent=parent,
                arguments=arguments,
                error=error,
            )
            if suppress_request and not is_interaction_exception:
                should_suppress = True
    
    return should_suppress


import inspect
import pickle

from typing import Callable, Dict, Optional, TYPE_CHECKING, Tuple, List, Any
from types import MethodType
from inspect import _ParameterKind
from typing_extensions import override
from bridgic.core.automa.worker import Worker
from bridgic.core.types._error import WorkerRuntimeError
from bridgic.core.utils._inspect_tools import load_qualified_class_or_func, get_param_names_all_kinds

if TYPE_CHECKING:
    from bridgic.core.automa._automa import Automa

class CallableWorker(Worker):
    """
    This class is a worker that wraps a callable object, such as functions or methods.

    Parameters
    ----------
    func_or_method : Optional[Callable]
        The callable to be wrapped by the worker. If `func_or_method` is None, 
        `state_dict` must be provided.
    """
    _is_async: bool
    _callable: Callable
    # Used to deserialization.
    _expected_bound_parent: bool

    # Cached method signatures, with no need for serialization.
    __cached_param_names_of_callable: Dict[_ParameterKind, List[str]]

    def __init__(
        self, 
        func_or_method: Optional[Callable] = None,
    ):
        """
        Parameters
        ----------
        func_or_method : Optional[Callable]
            The callable to be wrapped by the worker. If `func_or_method` is None, 
            `state_dict` must be provided.
        """
        super().__init__()
        self._is_async = inspect.iscoroutinefunction(func_or_method)
        self._callable = func_or_method
        self._expected_bound_parent = False
        
        # Cached method signatures, with no need for serialization.
        self.__cached_param_names_of_callable = None

    async def arun(self, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> Any:
        if self._expected_bound_parent:
            raise WorkerRuntimeError(
                f"The callable is expected to be bound to the parent, "
                f"but not bounded yet: {self._callable}"
            )
        if self._is_async:
            return await self._callable(*args, **kwargs)
        return await super().arun(*args, **kwargs)

    def run(self, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> Any:
        assert self._is_async is False
        return self._callable(*args, **kwargs)

    @override
    def get_input_param_names(self) -> Dict[_ParameterKind, List[str]]:
        """
        Get the names of input parameters of this callable worker.
        Use cached result if available in order to improve performance.

        Returns
        -------
        Dict[_ParameterKind, List[str]]
            A dictionary of input parameter names by the kind of the parameter.
            The key is the kind of the parameter, which is one of five possible values:

            - inspect.Parameter.POSITIONAL_ONLY
            - inspect.Parameter.POSITIONAL_OR_KEYWORD
            - inspect.Parameter.VAR_POSITIONAL
            - inspect.Parameter.KEYWORD_ONLY
            - inspect.Parameter.VAR_KEYWORD
        """
        if self.__cached_param_names_of_callable is None:
            self.__cached_param_names_of_callable = get_param_names_all_kinds(self._callable)
        return self.__cached_param_names_of_callable

    @override
    def dump_to_dict(self) -> Dict[str, Any]:
        state_dict = super().dump_to_dict()
        state_dict["is_async"] = self._is_async
        # Note: Not to use pickle to serialize the callable here.
        # We customize the serialization method of the callable to avoid creating instance multiple times and to minimize side effects.
        bounded = isinstance(self._callable, MethodType)
        state_dict["bounded"] = bounded
        if bounded:
            if self._callable.__self__ is self.parent:
                state_dict["callable_name"] = self._callable.__module__ + "." + self._callable.__qualname__
            else:
                state_dict["pickled_callable"] = pickle.dumps(self._callable)
        else:
            state_dict["callable_name"] = self._callable.__module__ + "." + self._callable.__qualname__
        return state_dict

    @override
    def load_from_dict(self, state_dict: Dict[str, Any]) -> None:
        super().load_from_dict(state_dict)
        # Deserialize from the state_dict.
        self._is_async = state_dict["is_async"]
        bounded = state_dict["bounded"]
        if bounded:
            pickled_callable = state_dict.get("pickled_callable", None)
            if pickled_callable is None:
                self._callable = load_qualified_class_or_func(state_dict["callable_name"])
                # Partially deserialized, need to be bound to the parent.
                self._expected_bound_parent = True
            else:
                self._callable = pickle.loads(pickled_callable)
                self._expected_bound_parent = False
        else:
            self._callable = load_qualified_class_or_func(state_dict["callable_name"])
            self._expected_bound_parent = False
        
        # Cached method signatures, with no need for serialization.
        self.__cached_param_names_of_callable = None

    @property
    def callable(self):
        return self._callable

    @property
    def parent(self) -> "Automa":
        return super().parent

    @parent.setter
    def parent(self, value: "Automa"):
        if self._expected_bound_parent:
            self._callable = MethodType(self._callable, value)
            self._expected_bound_parent = False
        Worker.parent.fset(self, value)

    @override
    def __str__(self) -> str:
        return f"CallableWorker(callable={self._callable.__name__})"
